# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import unittest

from . import FunctionalTestCase


class CleanupTest(FunctionalTestCase):
    u"""
    Test cleanup using duplicity binary
    """
    def test_cleanup_after_partial(self):
        u"""
        Regression test for https://bugs.launchpad.net/bugs/409593
        where duplicity deletes all the signatures during a cleanup
        after a failed backup.
        """
        self.make_largefiles()
        good_files = self.backup(u"full", u"testfiles/largefiles")
        good_files |= self.backup(u"inc", u"testfiles/largefiles")
        good_files |= self.backup(u"inc", u"testfiles/largefiles")
        self.backup(u"full", u"testfiles/largefiles", fail=1)
        bad_files = self.get_backend_files()
        bad_files -= good_files
        self.assertNotEqual(bad_files, set())
        # the cleanup should go OK
        self.run_duplicity(options=[u"cleanup", self.backend_url, u"--force"])
        leftovers = self.get_backend_files()
        self.assertEqual(good_files, leftovers)
        self.backup(u"inc", u"testfiles/largefiles")
        self.verify(u"testfiles/largefiles")

    def test_remove_all_but_n(self):
        u"""
        Test that remove-all-but-n works in the simple case.
        """
        full1_files = self.backup(u"full", u"testfiles/empty_dir")
        full2_files = self.backup(u"full", u"testfiles/empty_dir")
        self.run_duplicity(options=[u"remove-all-but-n", u"1", self.backend_url, u"--force"])
        leftovers = self.get_backend_files()
        self.assertEqual(full2_files, leftovers)

    def test_remove_all_inc_of_but_n(self):
        u"""
        Test that remove-all-inc-of-but-n-full works in the simple case.
        """
        full1_files = self.backup(u"full", u"testfiles/empty_dir")
        inc1_files = self.backup(u"inc", u"testfiles/empty_dir")
        full2_files = self.backup(u"full", u"testfiles/empty_dir")
        self.run_duplicity(options=[u"remove-all-inc-of-but-n-full", u"1", self.backend_url, u"--force"])
        leftovers = self.get_backend_files()
        self.assertEqual(full1_files | full2_files, leftovers)

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2012 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import os
import sys
import time
import unittest

from duplicity import backend
from duplicity import globals
from duplicity import log

_testing_dir = os.path.dirname(os.path.abspath(__file__))
_top_dir = os.path.dirname(_testing_dir)
_overrides_dir = os.path.join(_testing_dir, u'overrides')
_bin_dir = os.path.join(_testing_dir, u'overrides', u'bin')

# Adjust python path for duplicity and override modules
sys.path = [_overrides_dir, _top_dir, _bin_dir] + sys.path

# Also set PYTHONPATH for any subprocesses
os.environ[u'PYTHONPATH'] = _overrides_dir + u":" + _top_dir + u":" + os.environ.get(u'PYTHONPATH', u'')

# And PATH for any subprocesses
os.environ[u'PATH'] = _bin_dir + u":" + os.environ.get(u'PATH', u'')

# Now set some variables that help standardize test behavior
os.environ[u'LANG'] = u''
os.environ[u'GNUPGHOME'] = os.path.join(_testing_dir, u'gnupg')

# bzr does not honor perms so fix the perms and avoid annoying error
os.system(u"chmod 700 %s" % os.path.join(_testing_dir, u'gnupg'))

# Standardize time
os.environ[u'TZ'] = u'US/Central'
time.tzset()


class DuplicityTestCase(unittest.TestCase):

    sign_key = u'839E6A2856538CCF'
    sign_passphrase = u'test'
    encrypt_key1 = u'839E6A2856538CCF'
    encrypt_key2 = u'453005CE9B736B2A'

    def setUp(self):
        super(DuplicityTestCase, self).setUp()
        self.savedEnviron = {}
        self.savedGlobals = {}

        log.setup()
        log.setverbosity(log.WARNING)
        self.set_global(u'print_statistics', 0)
        backend.import_backends()

        # Have all file references in tests relative to our testing dir
        os.chdir(_testing_dir)

    def tearDown(self):
        for key in self.savedEnviron:
            self._update_env(key, self.savedEnviron[key])
        for key in self.savedGlobals:
            setattr(globals, key, self.savedGlobals[key])
        assert not os.system(u"rm -rf testfiles")
        super(DuplicityTestCase, self).tearDown()

    def unpack_testfiles(self):
        assert not os.system(u"rm -rf testfiles")
        assert not os.system(u"tar xzf testfiles.tar.gz > /dev/null 2>&1")
        assert not os.system(u"mkdir testfiles/output testfiles/cache")

    def _update_env(self, key, value):
        if value is not None:
            os.environ[key] = value
        elif key in os.environ:
            del os.environ[key]

    def set_environ(self, key, value):
        if key not in self.savedEnviron:
            self.savedEnviron[key] = os.environ.get(key)
        self._update_env(key, value)

    def set_global(self, key, value):
        assert hasattr(globals, key)
        if key not in self.savedGlobals:
            self.savedGlobals[key] = getattr(globals, key)
        setattr(globals, key, value)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2013 Matthieu Huin <mhu@enovance.com>
# Copyright 2015 Scott McKenzie <noizyland@gmail.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from builtins import str
import os

import duplicity.backend
from duplicity import globals
from duplicity import log
from duplicity.errors import BackendException
from duplicity.util import fsdecode


class AzureBackend(duplicity.backend.Backend):
    u"""
    Backend for Azure Blob Storage Service
    """
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Import Microsoft Azure Storage SDK for Python library.
        try:
            import azure
            import azure.storage
            if hasattr(azure.storage, u'BlobService'):
                # v0.11.1 and below
                from azure.storage import BlobService
                self.AzureMissingResourceError = azure.WindowsAzureMissingResourceError
                self.AzureConflictError = azure.WindowsAzureConflictError
            else:
                # v1.0.0 and above
                import azure.storage.blob
                if hasattr(azure.storage.blob, u'BlobService'):
                    from azure.storage.blob import BlobService
                else:
                    from azure.storage.blob.blockblobservice import BlockBlobService as BlobService
                self.AzureMissingResourceError = azure.common.AzureMissingResourceHttpError
                self.AzureConflictError = azure.common.AzureConflictHttpError
        except ImportError as e:
            raise BackendException(u"""\
Azure backend requires Microsoft Azure Storage SDK for Python (https://pypi.python.org/pypi/azure-storage/).
Exception: %s""" % str(e))

        # TODO: validate container name
        self.container = parsed_url.path.lstrip(u'/')

        if u'AZURE_ACCOUNT_NAME' not in os.environ:
            raise BackendException(u'AZURE_ACCOUNT_NAME environment variable not set.')

        if u'AZURE_ACCOUNT_KEY' in os.environ:
            if u'AZURE_ENDPOINT_SUFFIX' in os.environ:
                self.blob_service = BlobService(account_name=os.environ[u'AZURE_ACCOUNT_NAME'],
                                                account_key=os.environ[u'AZURE_ACCOUNT_KEY'],
                                                endpoint_suffix=os.environ[u'AZURE_ENDPOINT_SUFFIX'])
            else:
                self.blob_service = BlobService(account_name=os.environ[u'AZURE_ACCOUNT_NAME'],
                                                account_key=os.environ[u'AZURE_ACCOUNT_KEY'])
            self._create_container()
        elif u'AZURE_SHARED_ACCESS_SIGNATURE' in os.environ:
            if u'AZURE_ENDPOINT_SUFFIX' in os.environ:
                self.blob_service = BlobService(account_name=os.environ[u'AZURE_ACCOUNT_NAME'],
                                                sas_token=os.environ[u'AZURE_SHARED_ACCESS_SIGNATURE'],
                                                endpoint_suffix=os.environ[u'AZURE_ENDPOINT_SUFFIX'])
            else:
                self.blob_service = BlobService(account_name=os.environ[u'AZURE_ACCOUNT_NAME'],
                                                sas_token=os.environ[u'AZURE_SHARED_ACCESS_SIGNATURE'])
        else:
            raise BackendException(
                u'Neither AZURE_ACCOUNT_KEY nor AZURE_SHARED_ACCESS_SIGNATURE environment variable not set.')

        if globals.azure_max_single_put_size:
            # check if we use azure-storage>=0.30.0
            try:
                _ = self.blob_service.MAX_SINGLE_PUT_SIZE
                self.blob_service.MAX_SINGLE_PUT_SIZE = globals.azure_max_single_put_size
            # fallback for azure-storage<0.30.0
            except AttributeError:
                self.blob_service._BLOB_MAX_DATA_SIZE = globals.azure_max_single_put_size

        if globals.azure_max_block_size:
            # check if we use azure-storage>=0.30.0
            try:
                _ = self.blob_service.MAX_BLOCK_SIZE
                self.blob_service.MAX_BLOCK_SIZE = globals.azure_max_block_size
            # fallback for azure-storage<0.30.0
            except AttributeError:
                self.blob_service._BLOB_MAX_CHUNK_DATA_SIZE = globals.azure_max_block_size

    def _create_container(self):
        try:
            self.blob_service.create_container(self.container, fail_on_exist=True)
        except self.AzureConflictError:
            # Indicates that the resource could not be created because it already exists.
            pass
        except Exception as e:
            log.FatalError(u"Could not create Azure container: %s"
                           % str(e.message).split(u'\n', 1)[0],
                           log.ErrorCode.connection_failed)

    def _put(self, source_path, remote_filename):
        remote_filename = fsdecode(remote_filename)
        kwargs = {}
        if globals.azure_max_connections:
            kwargs[u'max_connections'] = globals.azure_max_connections

        # https://azure.microsoft.com/en-us/documentation/articles/storage-python-how-to-use-blob-storage/#upload-a-blob-into-a-container
        try:
            self.blob_service.create_blob_from_path(self.container, remote_filename, source_path.name, **kwargs)
        except AttributeError:  # Old versions use a different method name
            self.blob_service.put_block_blob_from_path(self.container, remote_filename, source_path.name, **kwargs)

        self._set_tier(remote_filename)

    def _set_tier(self, remote_filename):
        if globals.azure_blob_tier is not None:
            try:
                self.blob_service.set_standard_blob_tier(self.container, remote_filename, globals.azure_blob_tier)
            except AttributeError:  # might not be available in old API
                pass

    def _get(self, remote_filename, local_path):
        # https://azure.microsoft.com/en-us/documentation/articles/storage-python-how-to-use-blob-storage/#download-blobs
        self.blob_service.get_blob_to_path(self.container, fsdecode(remote_filename), local_path.name)

    def _list(self):
        # https://azure.microsoft.com/en-us/documentation/articles/storage-python-how-to-use-blob-storage/#list-the-blobs-in-a-container
        blobs = []
        marker = None
        while True:
            batch = self.blob_service.list_blobs(self.container, marker=marker)
            blobs.extend(batch)
            if not batch.next_marker:
                break
            marker = batch.next_marker
        return [blob.name for blob in blobs]

    def _delete(self, filename):
        # http://azure.microsoft.com/en-us/documentation/articles/storage-python-how-to-use-blob-storage/#delete-blobs
        self.blob_service.delete_blob(self.container, fsdecode(filename))

    def _query(self, filename):
        prop = self.blob_service.get_blob_properties(self.container, fsdecode(filename))
        try:
            info = {u'size': int(prop.properties.content_length)}
        except AttributeError:
            # old versions directly returned the properties
            info = {u'size': int(prop[u'content-length'])}
        return info

    def _error_code(self, operation, e):
        if isinstance(e, self.AzureMissingResourceError):
            return log.ErrorCode.backend_not_found


duplicity.backend.register_backend(u'azure', AzureBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import range
from builtins import object
from past.utils import old_div

import os
import platform
import pytest
import random
import sys
import unittest

from duplicity import gpg
from duplicity import path
from . import UnitTestCase


@pytest.mark.usefixtures(u"redirect_stdin")
class GPGTest(UnitTestCase):
    u"""Test GPGFile"""
    def setUp(self):
        super(GPGTest, self).setUp()
        self.unpack_testfiles()
        self.default_profile = gpg.GPGProfile(passphrase=u"foobar")

    def gpg_cycle(self, s, profile=None):
        u"""Test encryption/decryption cycle on string s"""
        epath = path.Path(u"testfiles/output/encrypted_file")
        if not profile:
            profile = self.default_profile
        encrypted_file = gpg.GPGFile(1, epath, profile)
        encrypted_file.write(s)
        encrypted_file.close()

        epath2 = path.Path(u"testfiles/output/encrypted_file")
        decrypted_file = gpg.GPGFile(0, epath2, profile)
        dec_buf = decrypted_file.read()
        decrypted_file.close()

        assert s == dec_buf, (len(s), len(dec_buf))

    def test_gpg1(self):
        u"""Test gpg short strings"""
        self.gpg_cycle(b"hello, world")
        self.gpg_cycle(b"ansoetuh aoetnuh aoenstuh aoetnuh asoetuh saoteuh ")

    def test_gpg2(self):
        u"""Test gpg long strings easily compressed"""
        self.gpg_cycle(b" " * 50000)
        self.gpg_cycle(b"aoeu" * 1000000)

    def test_gpg3(self):
        u"""Test on random data - must have /dev/urandom device"""
        infp = open(u"/dev/urandom", u"rb")
        rand_buf = infp.read(120000)
        infp.close()
        self.gpg_cycle(rand_buf)

    def test_gpg_asym(self):
        u"""Test GPG asymmetric encryption"""
        profile = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                 recipients=[self.encrypt_key1,
                                             self.encrypt_key2])
        self.gpg_cycle(b"aoensutha aonetuh saoe", profile)

        profile2 = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                  recipients=[self.encrypt_key1])
        self.gpg_cycle(b"aoeu" * 10000, profile2)

    def test_gpg_hidden_asym(self):
        u"""Test GPG asymmetric encryption with hidden key id"""
        profile = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                 hidden_recipients=[self.encrypt_key1,
                                                    self.encrypt_key2])
        self.gpg_cycle(b"aoensutha aonetuh saoe", profile)

        profile2 = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                  hidden_recipients=[self.encrypt_key1])
        self.gpg_cycle(b"aoeu" * 10000, profile2)

    def test_gpg_signing(self):
        u"""Test to make sure GPG reports the proper signature key"""
        plaintext = b"hello" * 50000

        signing_profile = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                         sign_key=self.sign_key,
                                         recipients=[self.encrypt_key1])

        epath = path.Path(u"testfiles/output/encrypted_file")
        encrypted_signed_file = gpg.GPGFile(1, epath, signing_profile)
        encrypted_signed_file.write(plaintext)
        encrypted_signed_file.close()

        decrypted_file = gpg.GPGFile(0, epath, signing_profile)
        assert decrypted_file.read() == plaintext
        decrypted_file.close()
        sig = decrypted_file.get_signature()
        assert sig == self.sign_key, sig

    def test_gpg_signing_and_hidden_encryption(self):
        u"""Test to make sure GPG reports the proper signature key even with hidden encryption key id"""
        plaintext = b"hello" * 50000

        signing_profile = gpg.GPGProfile(passphrase=self.sign_passphrase,
                                         sign_key=self.sign_key,
                                         hidden_recipients=[self.encrypt_key1])

        epath = path.Path(u"testfiles/output/encrypted_file")
        encrypted_signed_file = gpg.GPGFile(1, epath, signing_profile)
        encrypted_signed_file.write(plaintext)
        encrypted_signed_file.close()

        decrypted_file = gpg.GPGFile(0, epath, signing_profile)
        assert decrypted_file.read() == plaintext
        decrypted_file.close()
        sig = decrypted_file.get_signature()
        assert sig == self.sign_key, sig

    @unittest.skipIf(u"ppc64el" in platform.machine(), u"Skip on ppc64el machines")
    def test_GPGWriteFile(self):
        u"""Test GPGWriteFile"""
        size = 400 * 1000
        gwfh = GPGWriteFile_Helper()
        profile = gpg.GPGProfile(passphrase=u"foobar")
        for i in range(10):  # @UnusedVariable
            gpg.GPGWriteFile(gwfh, u"testfiles/output/gpgwrite.gpg",
                             profile, size=size)
            # print os.stat("testfiles/output/gpgwrite.gpg").st_size-size
            assert size - 64 * 1024 <= os.stat(u"testfiles/output/gpgwrite.gpg").st_size <= size + 64 * 1024
        gwfh.set_at_end()
        gpg.GPGWriteFile(gwfh, u"testfiles/output/gpgwrite.gpg",
                         profile, size=size)
        # print os.stat("testfiles/output/gpgwrite.gpg").st_size

    def test_GzipWriteFile(self):
        u"""Test GzipWriteFile"""
        size = 400 * 1000
        gwfh = GPGWriteFile_Helper()
        for i in range(10):  # @UnusedVariable
            gpg.GzipWriteFile(gwfh, u"testfiles/output/gzwrite.gz",
                              size=size)
            # print os.stat("testfiles/output/gzwrite.gz").st_size-size
            assert size - 64 * 1024 <= os.stat(u"testfiles/output/gzwrite.gz").st_size <= size + 64 * 1024
        gwfh.set_at_end()
        gpg.GzipWriteFile(gwfh, u"testfiles/output/gzwrite.gz", size=size)
        # print os.stat("testfiles/output/gzwrite.gz").st_size


class GPGWriteHelper2(object):
    def __init__(self, data):
        self.data = data


class GPGWriteFile_Helper(object):
    u"""Used in test_GPGWriteFile above"""
    def __init__(self):
        self.from_random_fp = open(u"/dev/urandom", u"rb")
        self.at_end = 0

    def set_at_end(self):
        u"""Iterator stops when you call this"""
        self.at_end = 1

    def get_buffer(self, size):
        u"""Return buffer of size size, consisting of half random data"""
        s1 = int(old_div(size, 2))
        s2 = size - s1
        return b"a" * s1 + self.from_random_fp.read(s2)

    def __next__(self):
        if self.at_end:
            raise StopIteration
        block_data = self.get_buffer(self.get_read_size())
        return GPGWriteHelper2(block_data)

    def get_read_size(self):
        size = 64 * 1024
        if random.randrange(2):
            return size
        else:
            return random.randrange(0, size)

    def get_footer(self):
        return b"e" * random.randrange(0, 15000)


class SHATest(UnitTestCase):
    u"""Test making sha signatures"""
    def setUp(self):
        super(SHATest, self).setUp()
        self.unpack_testfiles()

    def test_sha(self):
        hash = gpg.get_hash(u"SHA1", path.Path(u"testfiles/various_file_types/regular_file"))
        assert hash == u"886d722999862724e1e62d0ac51c468ee336ef8e", hash


if __name__ == u"__main__":
    unittest.main()



#
# Copyright (c) 2015 Matthew Bentley
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from builtins import object
from future import standard_library
standard_library.install_aliases()

import os
import hashlib
from urllib.parse import quote_plus  # pylint: disable=import-error

import duplicity.backend
from duplicity.errors import BackendException, FatalBackendException
from duplicity import log
from duplicity import progress
from duplicity import util


class B2ProgressListener(object):
    def __enter__(self):
        pass

    def set_total_bytes(self, total_byte_count):
        self.total_byte_count = total_byte_count

    def bytes_completed(self, byte_count):
        progress.report_transfer(byte_count, self.total_byte_count)

    def close(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class B2Backend(duplicity.backend.Backend):
    u"""
    Backend for BackBlaze's B2 storage service
    """

    def __init__(self, parsed_url):
        u"""
        Authorize to B2 api and set up needed variables
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        global DownloadDestLocalFile, FileVersionInfoFactory
        try:  # try to import the new b2sdk if available
            from b2sdk.api import B2Api
            from b2sdk.account_info import InMemoryAccountInfo
            from b2sdk.download_dest import DownloadDestLocalFile
            from b2sdk.exception import NonExistentBucket
            from b2sdk.file_version import FileVersionInfoFactory
        except ImportError:
            try:  # fall back to import the old b2 client
                from b2.api import B2Api
                from b2.account_info import InMemoryAccountInfo
                from b2.download_dest import DownloadDestLocalFile
                from b2.exception import NonExistentBucket
                from b2.file_version import FileVersionInfoFactory
            except ImportError:
                raise BackendException(u'B2 backend requires B2 Python SDK (pip install b2sdk)')

        self.service = B2Api(InMemoryAccountInfo())
        self.parsed_url.hostname = u'B2'

        account_id = parsed_url.username
        account_key = self.get_password()

        self.url_parts = [
            x for x in parsed_url.path.replace(u"@", u"/").split(u'/') if x != u''
        ]
        if self.url_parts:
            self.username = self.url_parts.pop(0)
            bucket_name = self.url_parts.pop(0)
        else:
            raise BackendException(u"B2 requires a bucket name")
        self.path = u"".join([url_part + u"/" for url_part in self.url_parts])
        self.service.authorize_account(u'production', account_id, account_key)

        log.Log(u"B2 Backend (path= %s, bucket= %s, minimum_part_size= %s)" %
                (self.path, bucket_name, self.service.account_info.get_minimum_part_size()), log.INFO)
        try:
            self.bucket = self.service.get_bucket_by_name(bucket_name)
            log.Log(u"Bucket found", log.INFO)
        except NonExistentBucket:
            try:
                log.Log(u"Bucket not found, creating one", log.INFO)
                self.bucket = self.service.create_bucket(bucket_name, u'allPrivate')
            except:
                raise FatalBackendException(u"Bucket cannot be created")

    def _get(self, remote_filename, local_path):
        u"""
        Download remote_filename to local_path
        """
        log.Log(u"Get: %s -> %s" % (self.path + util.fsdecode(remote_filename),
                                    util.fsdecode(local_path.name)),
                log.INFO)
        self.bucket.download_file_by_name(quote_plus(self.path + util.fsdecode(remote_filename), u'/'),
                                          DownloadDestLocalFile(util.fsdecode(local_path.name)))

    def _put(self, source_path, remote_filename):
        u"""
        Copy source_path to remote_filename
        """
        log.Log(u"Put: %s -> %s" % (util.fsdecode(source_path.name),
                                    self.path + util.fsdecode(remote_filename)),
                log.INFO)
        self.bucket.upload_local_file(util.fsdecode(source_path.name),
                                      quote_plus(self.path + util.fsdecode(remote_filename), u'/'),
                                      content_type=u'application/pgp-encrypted',
                                      progress_listener=B2ProgressListener())

    def _list(self):
        u"""
        List files on remote server
        """
        return [file_version_info.file_name[len(self.path):]
                for (file_version_info, folder_name) in self.bucket.ls(self.path)]

    def _delete(self, filename):
        u"""
        Delete filename from remote server
        """
        log.Log(u"Delete: %s" % self.path + util.fsdecode(filename), log.INFO)
        file_version_info = self.file_info(quote_plus(self.path + util.fsdecode(filename), u'/'))
        self.bucket.delete_file_version(file_version_info.id_, file_version_info.file_name)

    def _query(self, filename):
        u"""
        Get size info of filename
        """
        log.Log(u"Query: %s" % self.path + util.fsdecode(filename), log.INFO)
        file_version_info = self.file_info(quote_plus(self.path + util.fsdecode(filename), u'/'))
        return {u'size': file_version_info.size
                if file_version_info is not None and file_version_info.size is not None else -1}

    def file_info(self, filename):
        response = self.bucket.api.session.list_file_names(self.bucket.id_, filename, 1, None)
        for entry in response[u'files']:
            file_version_info = FileVersionInfoFactory.from_api_response(entry)
            if file_version_info.file_name == filename:
                return file_version_info
        raise BackendException(u'File not found')


duplicity.backend.register_backend(u"b2", B2Backend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2008 Peter Schuller <peter.schuller@infidyne.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""
Asynchronous job scheduler, for concurrent execution with minimalistic
dependency guarantees.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import object
import duplicity
from duplicity import log
from duplicity.dup_threading import require_threading
from duplicity.dup_threading import interruptably_wait
from duplicity.dup_threading import async_split
from duplicity.dup_threading import with_lock

thread = duplicity.dup_threading.thread_module()
threading = duplicity.dup_threading.threading_module()


class AsyncScheduler(object):
    u"""
    Easy-to-use scheduler of function calls to be executed
    concurrently. A very simple dependency mechanism exists in the
    form of barriers (see insert_barrier()).

    Each instance has a concurrency level associated with it. A
    concurrency of 0 implies that all tasks will be executed
    synchronously when scheduled. A concurrency of 1 indicates that a
    task will be executed asynchronously, but never concurrently with
    other tasks. Both 0 and 1 guarantee strict ordering among all
    tasks (i.e., they will be executed in the order scheduled).

    At concurrency levels above 1, the tasks will end up being
    executed in an order undetermined except insofar as is enforced by
    calls to insert_barrier().

    An AsynchScheduler should be created for any independent process;
    the scheduler will assume that if any background job fails (raises
    an exception), it makes further work moot.
    """

    def __init__(self, concurrency):
        u"""
        Create an asynchronous scheduler that executes jobs with the
        given level of concurrency.
        """
        log.Info(u"%s: %s" % (self.__class__.__name__,
                              _(u"instantiating at concurrency %d") %
                              (concurrency)))
        assert concurrency >= 0, u"%s concurrency level must be >= 0" % (self.__class__.__name__,)

        self.__failed = False  # has at least one task failed so far?
        self.__failed_waiter = None  # when __failed, the waiter of the first task that failed
        self.__concurrency = concurrency
        self.__worker_count = 0  # number of active workers
        self.__waiter_count = 0  # number of threads waiting to submit work
        self.__barrier = False  # barrier currently in effect?
        self.__cv = threading.Condition()  # for simplicity, we use a single cv with its lock
#                                                    # for everything, even if the resulting notifyAll():s
#                                                    # are not technically efficient.

        if concurrency > 0:
            require_threading(u"concurrency > 0 (%d)" % (concurrency,))

    def insert_barrier(self):
        u"""
        Proclaim that any tasks scheduled prior to the call to this
        method MUST be executed prior to any tasks scheduled after the
        call to this method.

        The intended use case is that if task B depends on A, a
        barrier must be inserted in between to guarantee that A
        happens before B.
        """
        log.Debug(u"%s: %s" % (self.__class__.__name__, _(u"inserting barrier")))
        # With concurrency 0 it's a NOOP, and due to the special case in
        # task scheduling we do not want to append to the queue (will never
        # be popped).
        if self.__concurrency > 0:
            def _insert_barrier():
                self.__barrier = True

            with_lock(self.__cv, _insert_barrier)

    def schedule_task(self, fn, params):
        u"""
        Schedule the given task (callable, typically function) for
        execution. Pass the given parameters to the function when
        calling it. Returns a callable which can optionally be used
        to wait for the task to complete, either by returning its
        return value or by propagating any exception raised by said
        task.

        This method may block or return immediately, depending on the
        configuration and state of the scheduler.

        This method may also raise an exception in order to trigger
        failures early, if the task (if run synchronously) or a previous
        task has already failed.

        NOTE: Pay particular attention to the scope in which this is
        called. In particular, since it will execute concurrently in
        the background, assuming fn is a closure, any variables used
        most be properly bound in the closure. This is the reason for
        the convenience feature of being able to give parameters to
        the call, to avoid having to wrap the call itself in a
        function in order to "fixate" variables in, for example, an
        enclosing loop.
        """
        assert fn is not None

        # Note: It is on purpose that we keep track of concurrency in
        # the front end and launch threads for each task, rather than
        # keep a pool of workers. The overhead is not relevant in the
        # situation this will be used, and it removes complexity in
        # terms of ensuring the scheduler is garbage collected/shut
        # down properly when no longer referenced/needed by calling
        # code.

        if self.__concurrency == 0:
            # special case this to not require any platform support for
            # threading at all
            log.Info(u"%s: %s" % (self.__class__.__name__,
                     _(u"running task synchronously (asynchronicity disabled)")),
                     log.InfoCode.synchronous_upload_begin)

            return self.__run_synchronously(fn, params)
        else:
            log.Info(u"%s: %s" % (self.__class__.__name__,
                     _(u"scheduling task for asynchronous execution")),
                     log.InfoCode.asynchronous_upload_begin)

            return self.__run_asynchronously(fn, params)

    def wait(self):
        u"""
        Wait for the scheduler to become entirely empty (i.e., all
        tasks having run to completion).

        IMPORTANT: This is only useful with a single caller scheduling
        tasks, such that no call to schedule_task() is currently in
        progress or may happen subsequently to the call to wait().
        """
        def _wait():
            interruptably_wait(self.__cv, lambda: self.__worker_count == 0 and self.__waiter_count == 0)

        with_lock(self.__cv, _wait)

    def __run_synchronously(self, fn, params):

        # When running synchronously, we immediately leak any exception raised
        # for immediate failure reporting to calling code.
        ret = fn(*params)

        def _waiter():
            return ret

        log.Info(u"%s: %s" % (self.__class__.__name__,
                 _(u"task completed successfully")),
                 log.InfoCode.synchronous_upload_done)

        return _waiter

    def __run_asynchronously(self, fn, params):
        (waiter, caller) = async_split(lambda: fn(*params))

        def check_pending_failure():
            if self.__failed:
                log.Info(u"%s: %s" % (self.__class__.__name__,
                         _(u"a previously scheduled task has failed; "
                           u"propagating the result immediately")),
                         log.InfoCode.asynchronous_upload_done)
                self.__failed_waiter()
                raise AssertionError(u"%s: waiter should have raised an exception; "
                                     u"this is a bug" % (self.__class__.__name__,))

        def wait_for_and_register_launch():
            check_pending_failure()  # raise on fail
            while self.__worker_count >= self.__concurrency or self.__barrier:
                if self.__worker_count == 0:
                    assert self.__barrier, u"barrier should be in effect"
                    self.__barrier = False
                    self.__cv.notifyAll()
                else:
                    self.__waiter_count += 1
                    self.__cv.wait()
                    self.__waiter_count -= 1

                check_pending_failure()  # raise on fail

            self.__worker_count += 1
            log.Debug(u"%s: %s" % (self.__class__.__name__,
                                   _(u"active workers = %d") % (self.__worker_count,)))

        # simply wait for an OK condition to start, then launch our worker. the worker
        # never waits on us, we just wait for them.
        with_lock(self.__cv, wait_for_and_register_launch)

        self.__start_worker(caller)

        return waiter

    def __start_worker(self, caller):
        u"""
        Start a new worker.
        """
        def trampoline():
            try:
                self.__execute_caller(caller)
            finally:
                def complete_worker():
                    self.__worker_count -= 1
                    log.Debug(u"%s: %s" % (self.__class__.__name__,
                                           _(u"active workers = %d") % (self.__worker_count,)))
                    self.__cv.notifyAll()
                with_lock(self.__cv, complete_worker)

        thread.start_new_thread(trampoline, ())

    def __execute_caller(self, caller):
        # The caller half that we get here will not propagate
        # errors back to us, but rather propagate it back to the
        # "other half" of the async split.
        succeeded, waiter = caller()
        if not succeeded:
            def _signal_failed():
                if not self.__failed:
                    self.__failed = True
                    self.__failed_waiter = waiter
                    self.__cv.notifyAll()
            with_lock(self.__cv, _signal_failed)

        log.Info(u"%s: %s" % (self.__class__.__name__,
                 _(u"task execution done (success: %s)") % succeeded),
                 log.InfoCode.asynchronous_upload_done)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Classes and functions on collections of backup volumes"""

from past.builtins import cmp
from builtins import filter
from builtins import str
from builtins import zip
from builtins import map
from builtins import range
from builtins import object

import types
import sys

from duplicity import log
from duplicity import file_naming
from duplicity import path
from duplicity import util
from duplicity import dup_time
from duplicity import globals
from duplicity import manifest
from duplicity import util
from duplicity.gpg import GPGError

# For type testing against both int and long types that works in python 2/3
if sys.version_info < (3,):
    integer_types = (int, int)
else:
    integer_types = (int,)


class CollectionsError(Exception):
    pass


class BackupSet(object):
    u"""
    Backup set - the backup information produced by one session
    """
    def __init__(self, backend, action):
        u"""
        Initialize new backup set, only backend is required at first
        """
        self.backend = backend
        self.info_set = False  # true if fields are set
        self.volume_name_dict = {}  # dict from volume number to filename
        self.remote_manifest_name = None  # full name of remote manifest
        self.local_manifest_path = None  # full path to local manifest
        self.time = None  # will be set if is full backup set
        self.start_time = None  # will be set if inc
        self.end_time = None  # will be set if inc
        self.partial = False  # true if a partial backup
        self.encrypted = False  # true if an encrypted backup
        self.files_changed = []
        self.action = action

    def is_complete(self):
        u"""
        Assume complete if found manifest file
        """
        return self.remote_manifest_name

    def add_filename(self, filename, pr=None):
        u"""
        Add a filename to given set.  Return true if it fits.

        The filename will match the given set if it has the right
        times and is of the right type.  The information will be set
        from the first filename given.

        @param filename: name of file to add
        @type filename: string

        @param pr: pre-computed result of file_naming.parse(filename)
        @type pr: Optional[ParseResults]
        """
        if not pr:
            pr = file_naming.parse(filename)
        if not pr or not (pr.type == u"full" or pr.type == u"inc"):
            return False

        if not self.info_set:
            self.set_info(pr)
        else:
            if pr.type != self.type:
                return False
            if pr.time != self.time:
                return False
            if (pr.start_time != self.start_time or
                    pr.end_time != self.end_time):
                return False
            if bool(pr.encrypted) != bool(self.encrypted):
                if self.partial and pr.encrypted:
                    self.encrypted = pr.encrypted

        if pr.manifest:
            self.set_manifest(filename)
        else:
            assert pr.volume_number is not None
            assert pr.volume_number not in self.volume_name_dict, \
                (self.volume_name_dict, filename)
            self.volume_name_dict[pr.volume_number] = filename

        return True

    def set_info(self, pr):
        u"""
        Set BackupSet information from ParseResults object

        @param pr: parse results
        @type pf: ParseResults
        """
        assert not self.info_set
        self.type = pr.type
        self.time = pr.time
        self.start_time = pr.start_time
        self.end_time = pr.end_time
        self.time = pr.time
        self.partial = pr.partial
        self.encrypted = bool(pr.encrypted)
        self.info_set = True

    def set_files_changed(self):
        mf = self.get_manifest()
        self.files_changed = mf.get_files_changed()

    def set_manifest(self, remote_filename):
        u"""
        Add local and remote manifest filenames to backup set
        """
        assert not self.remote_manifest_name, (self.remote_manifest_name,
                                               remote_filename)
        self.remote_manifest_name = remote_filename

        if self.action not in [u"collection-status", u"replicate"]:
            local_filename_list = globals.archive_dir_path.listdir()
        else:
            local_filename_list = []
        for local_filename in local_filename_list:
            pr = file_naming.parse(local_filename)
            if (pr and pr.manifest and pr.type == self.type and
                    pr.time == self.time and
                    pr.start_time == self.start_time and
                    pr.end_time == self.end_time):
                self.local_manifest_path = \
                    globals.archive_dir_path.append(local_filename)

                self.set_files_changed()
                break

    def delete(self):
        u"""
        Remove all files in set, both local and remote
        """
        rfn = self.get_filenames()
        rfn.reverse()
        try:
            self.backend.delete(rfn)
        except Exception:
            log.Debug(_(u"BackupSet.delete: missing %s") % [util.fsdecode(f) for f in rfn])
            pass
        if self.action not in [u"collection-status", u"replicate"]:
            local_filename_list = globals.archive_dir_path.listdir()
        else:
            local_filename_list = []
        for lfn in local_filename_list:
            pr = file_naming.parse(lfn)
            if (pr and pr.time == self.time and
                    pr.start_time == self.start_time and
                    pr.end_time == self.end_time):
                try:
                    globals.archive_dir_path.append(lfn).delete()
                except Exception:
                    log.Debug(_(u"BackupSet.delete: missing %s") % [util.fsdecode(f) for f in lfn])
                    pass
        util.release_lockfile()

    def __str__(self):
        u"""
        For now just list files in set
        """
        filelist = []
        if self.remote_manifest_name:
            filelist.append(self.remote_manifest_name)
        filelist.extend(list(self.volume_name_dict.values()))
        return u"[%s]" % u", ".join(map(util.fsdecode, filelist))

    def get_timestr(self):
        u"""
        Return time string suitable for log statements
        """
        return dup_time.timetopretty(self.time or self.end_time)

    def check_manifests(self, check_remote=True):
        u"""
        Make sure remote manifest is equal to local one
        """
        if not self.remote_manifest_name and not self.local_manifest_path:
            log.FatalError(_(u"Fatal Error: No manifests found for most recent backup"),
                           log.ErrorCode.no_manifests)
        assert self.remote_manifest_name, u"if only one, should be remote"

        remote_manifest = self.get_remote_manifest() if check_remote else None
        if self.local_manifest_path:
            local_manifest = self.get_local_manifest()
        if remote_manifest and self.local_manifest_path and local_manifest:
            if remote_manifest != local_manifest:
                log.FatalError(_(u"Fatal Error: Remote manifest does not match "
                                 u"local one.  Either the remote backup set or "
                                 u"the local archive directory has been corrupted."),
                               log.ErrorCode.mismatched_manifests)
        if not remote_manifest:
            if self.local_manifest_path:
                remote_manifest = local_manifest
            else:
                log.FatalError(_(u"Fatal Error: Neither remote nor local "
                                 u"manifest is readable."),
                               log.ErrorCode.unreadable_manifests)
        remote_manifest.check_dirinfo()

    def get_local_manifest(self):
        u"""
        Return manifest object by reading local manifest file
        """
        assert self.local_manifest_path
        manifest_buffer = self.local_manifest_path.get_data()
        log.Info(_(u"Processing local manifest %s (%s)") % (
            self.local_manifest_path.name, len(manifest_buffer)))
        return manifest.Manifest().from_string(manifest_buffer)

    def get_remote_manifest(self):
        u"""
        Return manifest by reading remote manifest on backend
        """
        assert self.remote_manifest_name
        try:
            manifest_buffer = self.backend.get_data(self.remote_manifest_name)
        except GPGError as message:
            log.Error(_(u"Error processing remote manifest (%s): %s") %
                      (util.fsdecode(self.remote_manifest_name), util.uexc(message)))
            return None
        log.Info(_(u"Processing remote manifest %s (%s)") % (
            util.fsdecode(self.remote_manifest_name), len(manifest_buffer)))
        return manifest.Manifest().from_string(manifest_buffer)

    def get_manifest(self):
        u"""
        Return manifest object, showing preference for local copy
        """
        if self.local_manifest_path:
            return self.get_local_manifest()
        else:
            return self.get_remote_manifest()

    def get_filenames(self):
        u"""
        Return sorted list of (remote) filenames of files in set
        """
        assert self.info_set
        volume_num_list = list(self.volume_name_dict.keys())
        volume_num_list.sort()
        volume_filenames = [self.volume_name_dict[x] for x in volume_num_list]
        if self.remote_manifest_name:
            # For convenience of implementation for restart support, we treat
            # local partial manifests as this set's remote manifest.  But
            # when specifically asked for a list of remote filenames, we
            # should not include it.
            pr = file_naming.parse(self.remote_manifest_name)
            if pr and not pr.partial:
                volume_filenames.append(self.remote_manifest_name)
        return volume_filenames

    def get_time(self):
        u"""
        Return time if full backup, or end_time if incremental
        """
        if self.time:
            return self.time
        if self.end_time:
            return self.end_time
        assert 0, u"Neither self.time nor self.end_time set"

    def get_files_changed(self):
        return self.files_changed

    def __len__(self):
        u"""
        Return the number of volumes in the set
        """
        return len(list(self.volume_name_dict.keys()))

    def __eq__(self, other):
        u"""
        Return whether this backup set is equal to other
        """
        return self.type == other.type and \
            self.time == other.time and \
            self.start_time == other.start_time and \
            self.end_time == other.end_time and \
            len(self) == len(other)


class BackupChain(object):
    u"""
    BackupChain - a number of linked BackupSets

    A BackupChain always starts with a full backup set and continues
    with incremental ones.
    """
    def __init__(self, backend):
        u"""
        Initialize new chain, only backend is required at first
        """
        self.backend = backend
        self.fullset = None
        self.incset_list = []  # sorted list of BackupSets
        self.start_time, self.end_time = None, None

    def set_full(self, fullset):
        u"""
        Add full backup set
        """
        assert not self.fullset and isinstance(fullset, BackupSet)
        self.fullset = fullset
        assert fullset.time
        self.start_time, self.end_time = fullset.time, fullset.time

    def add_inc(self, incset):
        u"""
        Add incset to self.  Return False if incset does not match
        """
        if self.end_time == incset.start_time:
            self.incset_list.append(incset)
        else:
            if (self.incset_list and
                    incset.start_time == self.incset_list[-1].start_time and
                    incset.end_time > self.incset_list[-1].end_time):
                log.Info(_(u"Preferring Backupset over previous one!"))
                self.incset_list[-1] = incset
            else:
                log.Info(_(u"Ignoring incremental Backupset (start_time: %s; needed: %s)") %
                         (dup_time.timetopretty(incset.start_time),
                          dup_time.timetopretty(self.end_time)))
                return False
        self.end_time = incset.end_time
        log.Info(_(u"Added incremental Backupset (start_time: %s / end_time: %s)") %
                 (dup_time.timetopretty(incset.start_time),
                  dup_time.timetopretty(incset.end_time)))
        assert self.end_time
        return True

    def delete(self, keep_full=False):
        u"""
        Delete all sets in chain, in reverse order
        """
        for i in range(len(self.incset_list) - 1, -1, -1):
            self.incset_list[i].delete()
        if self.fullset and not keep_full:
            self.fullset.delete()

    def get_sets_at_time(self, time):
        u"""
        Return a list of sets in chain earlier or equal to time
        """
        older_incsets = [s for s in self.incset_list if s.end_time <= time]
        return [self.fullset] + older_incsets

    def get_last(self):
        u"""
        Return last BackupSet in chain
        """
        if self.incset_list:
            return self.incset_list[-1]
        else:
            return self.fullset

    def get_first(self):
        u"""
        Return first BackupSet in chain (ie the full backup)
        """
        return self.fullset

    def short_desc(self):
        u"""
        Return a short one-line description of the chain,
        suitable for log messages.
        """
        return u"[%s]-[%s]" % (dup_time.timetopretty(self.start_time),
                               dup_time.timetopretty(self.end_time))

    def to_log_info(self, prefix=u''):
        u"""
        Return summary, suitable for printing to log
        """
        l = []
        for s in self.get_all_sets():
            if s.time:
                type = u"full"
                time = s.time
            else:
                type = u"inc"
                time = s.end_time
            if s.encrypted:
                enc = u"enc"
            else:
                enc = u"noenc"
            l.append(u"%s%s %s %d %s" % (prefix, type, dup_time.timetostring(time), (len(s)), enc))
        return l

    def __str__(self):
        u"""
        Return string representation, for testing purposes
        """
        set_schema = u"%20s   %30s   %15s"
        l = [u"-------------------------",
             _(u"Chain start time: ") + dup_time.timetopretty(self.start_time),
             _(u"Chain end time: ") + dup_time.timetopretty(self.end_time),
             _(u"Number of contained backup sets: %d") %
             (len(self.incset_list) + 1,),
             _(u"Total number of contained volumes: %d") %
             (self.get_num_volumes(),),
             set_schema % (_(u"Type of backup set:"), _(u"Time:"), _(u"Num volumes:"))]

        for s in self.get_all_sets():
            if s.time:
                type = _(u"Full")
                time = s.time
            else:
                type = _(u"Incremental")
                time = s.end_time
            l.append(set_schema % (type, dup_time.timetopretty(time), len(s)))

        l.append(u"-------------------------")
        return u"\n".join(l)

    def get_num_volumes(self):
        u"""
        Return the total number of volumes in the chain
        """
        n = 0
        for s in self.get_all_sets():
            n += len(s)
        return n

    def get_all_sets(self):
        u"""
        Return list of all backup sets in chain
        """
        if self.fullset:
            return [self.fullset] + self.incset_list
        else:
            return self.incset_list


class SignatureChain(object):
    u"""
    A number of linked SignatureSets

    Analog to BackupChain - start with a full-sig, and continue with
    new-sigs.
    """
    def __init__(self, local, location):
        u"""
        Return new SignatureChain.

        local should be true iff the signature chain resides in
        globals.archive_dir_path and false if the chain is in
        globals.backend.

        @param local: True if sig chain in globals.archive_dir_path
        @type local: Boolean

        @param location: Where the sig chain is located
        @type location: globals.archive_dir_path or globals.backend
        """
        if local:
            self.archive_dir_path, self.backend = location, None
        else:
            self.archive_dir_path, self.backend = None, location
        self.fullsig = None  # filename of full signature
        self.inclist = []  # list of filenames of incremental signatures
        self.start_time, self.end_time = None, None

    def __str__(self):
        u"""
        Local or Remote and List of files in the set
        """
        if self.archive_dir_path:
            place = _(u"local")
        else:
            place = _(u"remote")
        filelist = []
        if self.fullsig:
            filelist.append(self.fullsig)
        filelist.extend(self.inclist)
        return u"%s: [%s]" % (place, u", ".join(filelist))

    def check_times(self, time_list):
        u"""
        Check to make sure times are in whole seconds
        """
        for time in time_list:
            if type(time) not in integer_types:
                assert 0, u"Time %s in %s wrong type" % (time, time_list)

    def islocal(self):
        u"""
        Return true if represents a signature chain in archive_dir_path
        """
        if self.archive_dir_path:
            return True
        else:
            return False

    def add_filename(self, filename, pr=None):
        u"""
        Add new sig filename to current chain.  Return true if fits
        """
        if not pr:
            pr = file_naming.parse(filename)
        if not pr:
            return None

        if self.fullsig:
            if pr.type != u"new-sig":
                return None
            if pr.start_time != self.end_time:
                return None
            self.inclist.append(filename)
            self.check_times([pr.end_time])
            self.end_time = pr.end_time
            return 1
        else:
            if pr.type != u"full-sig":
                return None
            self.fullsig = filename
            self.check_times([pr.time, pr.time])
            self.start_time, self.end_time = pr.time, pr.time
            return 1

    def get_fileobjs(self, time=None):
        u"""
        Return ordered list of signature fileobjs opened for reading,
        optionally at a certain time
        """
        assert self.fullsig
        if self.archive_dir_path:  # local
            def filename_to_fileobj(filename):
                u"""Open filename in archive_dir_path, return filtered fileobj"""
                sig_dp = path.DupPath(self.archive_dir_path.name, (filename,))
                return sig_dp.filtered_open(u"rb")
        else:
            filename_to_fileobj = self.backend.get_fileobj_read
        return [filename_to_fileobj(f) for f in self.get_filenames(time)]

    def delete(self, keep_full=False):
        u"""
        Remove all files in signature set
        """
        # Try to delete in opposite order, so something useful even if aborted
        if self.archive_dir_path:
            for i in range(len(self.inclist) - 1, -1, -1):
                self.archive_dir_path.append(self.inclist[i]).delete()
            if not keep_full:
                self.archive_dir_path.append(self.fullsig).delete()
        else:
            assert self.backend
            inclist_copy = self.inclist[:]
            inclist_copy.reverse()
            if not keep_full:
                inclist_copy.append(self.fullsig)
            self.backend.delete(inclist_copy)

    def get_filenames(self, time=None):
        u"""
        Return ordered list of filenames in set, up to a provided time
        """
        if self.fullsig:
            l = [self.fullsig]
        else:
            l = []

        inclist = self.inclist
        if time:
            inclist = [n for n in inclist if file_naming.parse(n).end_time <= time]

        l.extend(inclist)
        return l


class CollectionsStatus(object):
    u"""
    Hold information about available chains and sets
    """
    def __init__(self, backend, archive_dir_path, action):
        u"""
        Make new object.  Does not set values
        """
        self.backend = backend
        self.archive_dir_path = archive_dir_path
        self.action = action

        # Will hold (signature chain, backup chain) pair of active
        # (most recent) chains
        self.matched_chain_pair = None

        # These should be sorted by end_time
        self.all_backup_chains = None
        self.other_backup_chains = None
        self.all_sig_chains = None

        # Other misc paths and sets which shouldn't be there
        self.local_orphaned_sig_names = []
        self.remote_orphaned_sig_names = []
        self.orphaned_backup_sets = None
        self.incomplete_backup_sets = None

        # True if set_values() below has run
        self.values_set = None

    def to_log_info(self):
        u"""
        Return summary of the collection, suitable for printing to log
        """
        l = [u"backend %s" % (self.backend.__class__.__name__,),
             u"archive-dir %s" % (self.archive_dir_path,)]

        for i in range(len(self.other_backup_chains)):
            # A bit of a misnomer.  Chain might have a sig.
            l.append(u"chain-no-sig %d" % (i,))
            l += self.other_backup_chains[i].to_log_info(u' ')

        if self.matched_chain_pair:
            l.append(u"chain-complete")
            l += self.matched_chain_pair[1].to_log_info(u' ')

        l.append(u"orphaned-sets-num %d" % (len(self.orphaned_backup_sets),))
        l.append(u"incomplete-sets-num %d" % (len(self.incomplete_backup_sets),))

        return l

    def __str__(self):
        u"""
        Return string summary of the collection
        """
        l = [_(u"Collection Status"),
             u"-----------------",
             _(u"Connecting with backend: %s") %
             (self.backend.__class__.__name__,),
             _(u"Archive dir: %s") % (self.archive_dir_path.uc_name if self.archive_dir_path else u'None',)]

        l.append(u"\n" +
                 ngettext(u"Found %d secondary backup chain.",
                          u"Found %d secondary backup chains.",
                          len(self.other_backup_chains))
                 % len(self.other_backup_chains))
        for i in range(len(self.other_backup_chains)):
            l.append(_(u"Secondary chain %d of %d:") %
                     (i + 1, len(self.other_backup_chains)))
            l.append(str(self.other_backup_chains[i]))
            l.append(u"")

        if self.matched_chain_pair:
            l.append(u"\n" + _(u"Found primary backup chain with matching "
                     u"signature chain:"))
            l.append(str(self.matched_chain_pair[1]))
        else:
            l.append(_(u"No backup chains with active signatures found"))

        if self.orphaned_backup_sets or self.incomplete_backup_sets:
            l.append(ngettext(u"Also found %d backup set not part of any chain,",
                              u"Also found %d backup sets not part of any chain,",
                              len(self.orphaned_backup_sets))
                     % (len(self.orphaned_backup_sets),))
            l.append(ngettext(u"and %d incomplete backup set.",
                              u"and %d incomplete backup sets.",
                              len(self.incomplete_backup_sets))
                     % (len(self.incomplete_backup_sets),))
            # TRANSL: "cleanup" is a hard-coded command, so do not translate it
            l.append(_(u'These may be deleted by running duplicity with the '
                       u'"cleanup" command.'))
        else:
            l.append(_(u"No orphaned or incomplete backup sets found."))

        return u"\n".join(l)

    def set_values(self, sig_chain_warning=1):
        u"""
        Set values from archive_dir_path and backend.

        Returns self for convenience.  If sig_chain_warning is set to None,
        do not warn about unnecessary sig chains.  This is because there may
        naturally be some unecessary ones after a full backup.
        """
        self.values_set = 1

        # get remote filename list
        backend_filename_list = self.backend.list()
        log.Debug(ngettext(u"%d file exists on backend",
                           u"%d files exist on backend",
                           len(backend_filename_list)) %
                  len(backend_filename_list))

        # get local filename list
        if self.action not in [u"collection-status", u"replicate"]:
            local_filename_list = self.archive_dir_path.listdir()
        else:
            local_filename_list = []
        log.Debug(ngettext(u"%d file exists in cache",
                           u"%d files exist in cache",
                           len(local_filename_list)) %
                  len(local_filename_list))

        # check for partial backups
        partials = []
        for local_filename in local_filename_list:
            pr = file_naming.parse(local_filename)
            if pr and pr.partial:
                partials.append(local_filename)

        # get various backup sets and chains
        (backup_chains, self.orphaned_backup_sets,
         self.incomplete_backup_sets) = \
            self.get_backup_chains(partials + backend_filename_list)
        backup_chains = self.get_sorted_chains(backup_chains)
        self.all_backup_chains = backup_chains

        assert len(backup_chains) == len(self.all_backup_chains), \
            u"get_sorted_chains() did something more than re-ordering"

        local_sig_chains, self.local_orphaned_sig_names = \
            self.get_signature_chains(True)
        remote_sig_chains, self.remote_orphaned_sig_names = \
            self.get_signature_chains(False, filelist=backend_filename_list)
        self.set_matched_chain_pair(local_sig_chains + remote_sig_chains,
                                    backup_chains)
        self.warn(sig_chain_warning)
        return self

    def set_matched_chain_pair(self, sig_chains, backup_chains):
        u"""
        Set self.matched_chain_pair and self.other_sig/backup_chains

        The latest matched_chain_pair will be set.  If there are both
        remote and local signature chains capable of matching the
        latest backup chain, use the local sig chain (it does not need
        to be downloaded).
        """
        sig_chains = sig_chains and self.get_sorted_chains(sig_chains)
        self.all_sig_chains = sig_chains
        self.other_backup_chains = backup_chains[:]
        self.matched_chain_pair = None
        if sig_chains and backup_chains:
            latest_backup_chain = backup_chains[-1]
            for i in range(len(sig_chains) - 1, -1, -1):
                if sig_chains[i].end_time == latest_backup_chain.end_time:
                    pass
                # See if the set before last matches:
                elif (len(latest_backup_chain.get_all_sets()) >= 2 and
                      sig_chains[i].end_time == latest_backup_chain.get_all_sets()[-2].end_time):
                    # It matches, remove the last backup set:
                    log.Warn(_(u"Warning, discarding last backup set, because "
                               u"of missing signature file."))
                    self.incomplete_backup_sets.append(latest_backup_chain.incset_list[-1])
                    latest_backup_chain.incset_list = latest_backup_chain.incset_list[:-1]
                else:
                    continue

                # Found a matching pair:
                if self.matched_chain_pair is None:
                    self.matched_chain_pair = (sig_chains[i], latest_backup_chain)

                break

        if self.matched_chain_pair:
            self.other_backup_chains.remove(self.matched_chain_pair[1])

    def warn(self, sig_chain_warning):
        u"""
        Log various error messages if find incomplete/orphaned files
        """
        assert self.values_set

        if self.local_orphaned_sig_names:
            log.Warn(ngettext(u"Warning, found the following local orphaned "
                              u"signature file:",
                              u"Warning, found the following local orphaned "
                              u"signature files:",
                              len(self.local_orphaned_sig_names)) + u"\n" +
                     u"\n".join(map(util.fsdecode, self.local_orphaned_sig_names)),
                     log.WarningCode.orphaned_sig)

        if self.remote_orphaned_sig_names:
            log.Warn(ngettext(u"Warning, found the following remote orphaned "
                              u"signature file:",
                              u"Warning, found the following remote orphaned "
                              u"signature files:",
                              len(self.remote_orphaned_sig_names)) + u"\n" +
                     u"\n".join(map(util.fsdecode, self.remote_orphaned_sig_names)),
                     log.WarningCode.orphaned_sig)

        if self.all_sig_chains and sig_chain_warning and not self.matched_chain_pair:
            log.Warn(_(u"Warning, found signatures but no corresponding "
                       u"backup files"), log.WarningCode.unmatched_sig)

        if self.incomplete_backup_sets:
            log.Warn(_(u"Warning, found incomplete backup sets, probably left "
                       u"from aborted session"), log.WarningCode.incomplete_backup)

        if self.orphaned_backup_sets:
            log.Warn(ngettext(u"Warning, found the following orphaned "
                              u"backup file:",
                              u"Warning, found the following orphaned "
                              u"backup files:",
                              len(self.orphaned_backup_sets)) + u"\n" +
                     u"\n".join(map(str, self.orphaned_backup_sets)),
                     log.WarningCode.orphaned_backup)

    def get_backup_chains(self, filename_list):
        u"""
        Split given filename_list into chains

        Return value will be tuple (list of chains, list of sets, list
        of incomplete sets), where the list of sets will comprise sets
        not fitting into any chain, and the incomplete sets are sets
        missing files.
        """
        log.Debug(_(u"Extracting backup chains from list of files: %s")
                  % [util.fsdecode(f) for f in filename_list])
        # First put filenames in set form
        sets = []

        def add_to_sets(filename):
            u"""
            Try adding filename to existing sets, or make new one
            """
            pr = file_naming.parse(filename)
            for set in sets:
                if set.add_filename(filename, pr):
                    log.Debug(_(u"File %s is part of known set") % (util.fsdecode(filename),))
                    break
            else:
                log.Debug(_(u"File %s is not part of a known set; creating new set") % (util.fsdecode(filename),))
                new_set = BackupSet(self.backend, self.action)
                if new_set.add_filename(filename, pr):
                    sets.append(new_set)
                else:
                    log.Debug(_(u"Ignoring file (rejected by backup set) '%s'") % util.fsdecode(filename))

        for f in filename_list:
            add_to_sets(f)
        sets, incomplete_sets = self.get_sorted_sets(sets)

        chains, orphaned_sets = [], []

        def add_to_chains(set):
            u"""
            Try adding set to existing chains, or make new one
            """
            if set.type == u"full":
                new_chain = BackupChain(self.backend)
                new_chain.set_full(set)
                chains.append(new_chain)
                log.Debug(_(u"Found backup chain %s") % (new_chain.short_desc()))
            else:
                assert set.type == u"inc"
                for chain in chains:
                    if chain.add_inc(set):
                        log.Debug(_(u"Added set %s to pre-existing chain %s") % (set.get_timestr(),
                                                                                 chain.short_desc()))
                        break
                else:
                    log.Debug(_(u"Found orphaned set %s") % (set.get_timestr(),))
                    orphaned_sets.append(set)
        for s in sets:
            add_to_chains(s)
        return (chains, orphaned_sets, incomplete_sets)

    def get_sorted_sets(self, set_list):
        u"""
        Sort set list by end time, return (sorted list, incomplete)
        """
        time_set_pairs, incomplete_sets = [], []
        for set in set_list:
            if not set.is_complete():
                incomplete_sets.append(set)
            elif set.type == u"full":
                time_set_pairs.append((set.time, set))
            else:
                time_set_pairs.append((set.end_time, set))
        time_set_pairs.sort()
        return ([p[1] for p in time_set_pairs], incomplete_sets)

    def get_signature_chains(self, local, filelist=None):
        u"""
        Find chains in archive_dir_path (if local is true) or backend

        Use filelist if given, otherwise regenerate.  Return value is
        pair (list of chains, list of signature paths not in any
        chains).
        """
        def get_filelist():
            if filelist is not None:
                return filelist
            elif local:
                if self.action not in [u"collection-status", u"replicate"]:
                    return self.archive_dir_path.listdir()
                else:
                    return []
            else:
                return self.backend.list()

        def get_new_sigchain():
            u"""
            Return new empty signature chain
            """
            if local:
                return SignatureChain(True, self.archive_dir_path)
            else:
                return SignatureChain(False, self.backend)

        # Build initial chains from full sig filenames
        chains, new_sig_filenames = [], []
        for filename in get_filelist():
            pr = file_naming.parse(filename)
            if pr:
                if pr.type == u"full-sig":
                    new_chain = get_new_sigchain()
                    assert new_chain.add_filename(filename, pr)
                    chains.append(new_chain)
                elif pr.type == u"new-sig":
                    new_sig_filenames.append(filename)

        # Try adding new signatures to existing chains
        orphaned_filenames = []
        new_sig_filenames.sort(key=lambda x: int(file_naming.parse(x).start_time))
        for sig_filename in new_sig_filenames:
            for chain in chains:
                if chain.add_filename(sig_filename):
                    break
            else:
                orphaned_filenames.append(sig_filename)
        return (chains, orphaned_filenames)

    def get_sorted_chains(self, chain_list):
        u"""
        Return chains sorted by end_time.  If tie, local goes last
        """
        # Build dictionary from end_times to lists of corresponding chains
        endtime_chain_dict = {}
        for chain in chain_list:
            if chain.end_time in endtime_chain_dict:
                endtime_chain_dict[chain.end_time].append(chain)
            else:
                endtime_chain_dict[chain.end_time] = [chain]

        # Use dictionary to build final sorted list
        sorted_end_times = list(endtime_chain_dict.keys())
        sorted_end_times.sort()
        sorted_chain_list = []
        for end_time in sorted_end_times:
            chain_list = endtime_chain_dict[end_time]
            if len(chain_list) == 1:
                sorted_chain_list.append(chain_list[0])
            else:
                assert len(chain_list) == 2
                if chain_list[0].backend:  # is remote, goes first
                    sorted_chain_list.append(chain_list[0])
                    sorted_chain_list.append(chain_list[1])
                else:  # is local, goes second
                    sorted_chain_list.append(chain_list[1])
                    sorted_chain_list.append(chain_list[0])

        return sorted_chain_list

    def get_backup_chain_at_time(self, time):
        u"""
        Return backup chain covering specified time

        Tries to find the backup chain covering the given time.  If
        there is none, return the earliest chain before, and failing
        that, the earliest chain.
        """
        if not self.all_backup_chains:
            raise CollectionsError(u"No backup chains found")

        covering_chains = [c for c in self.all_backup_chains
                           if c.start_time <= time <= c.end_time]
        if len(covering_chains) > 1:
            raise CollectionsError(u"Two chains cover the given time")
        elif len(covering_chains) == 1:
            return covering_chains[0]

        old_chains = [c for c in self.all_backup_chains if c.end_time < time]
        if old_chains:
            return old_chains[-1]
        else:
            return self.all_backup_chains[0]  # no chains are old enough

    def get_signature_chain_at_time(self, time):
        u"""
        Return signature chain covering specified time

        Tries to find the signature chain covering the given time.  If
        there is none, return the earliest chain before, and failing
        that, the earliest chain.
        """
        if not self.all_sig_chains:
            raise CollectionsError(u"No signature chains found")

        covering_chains = [c for c in self.all_sig_chains
                           if c.start_time <= time <= c.end_time]
        if covering_chains:
            return covering_chains[-1]  # prefer local if multiple sig chains

        old_chains = [c for c in self.all_sig_chains if c.end_time < time]
        if old_chains:
            return old_chains[-1]
        else:
            # no chains are old enough, give oldest and warn user
            oldest = self.all_sig_chains[0]
            if time < oldest.start_time:
                log.Warn(_(u"No signature chain for the requested time. "
                           u"Using oldest available chain, starting at time %s.") %
                         dup_time.timetopretty(oldest.start_time),
                         log.WarningCode.no_sig_for_time,
                         dup_time.timetostring(oldest.start_time))
            return oldest

    def get_extraneous(self, extra_clean):
        u"""
        Return list of the names of extraneous duplicity files

        A duplicity file is considered extraneous if it is
        recognizable as a duplicity file, but isn't part of some
        complete backup set, or current signature chain.
        """
        assert self.values_set
        local_filenames = []
        remote_filenames = []
        ext_containers = self.orphaned_backup_sets + self.incomplete_backup_sets
        if extra_clean:
            old_sig_chains = self.all_sig_chains[:]
            if self.matched_chain_pair:
                matched_sig_chain = self.matched_chain_pair[0]
                for sig_chain in self.all_sig_chains:
                    if (sig_chain.start_time == matched_sig_chain.start_time and
                            sig_chain.end_time == matched_sig_chain.end_time):
                        old_sig_chains.remove(sig_chain)
            ext_containers += old_sig_chains
        for set_or_chain in ext_containers:
            if set_or_chain.backend:
                remote_filenames.extend(set_or_chain.get_filenames())
            else:
                local_filenames.extend(set_or_chain.get_filenames())
        local_filenames += self.local_orphaned_sig_names
        remote_filenames += self.remote_orphaned_sig_names
        return local_filenames, remote_filenames

    def sort_sets(self, setlist):
        u"""Return new list containing same elems of setlist, sorted by time"""
        pairs = [(s.get_time(), s) for s in setlist]
        pairs.sort()
        return [p[1] for p in pairs]

    def get_chains_older_than(self, t):
        u"""
        Returns a list of backup chains older than the given time t

        All of the times will be associated with an intact chain.
        Furthermore, none of the times will be of a chain which a newer
        set may depend on.  For instance, if set A is a full set older
        than t, and set B is an incremental based on A which is newer
        than t, then the time of set A will not be returned.
        """
        assert self.values_set
        old_chains = []
        for chain in self.all_backup_chains:
            if (chain.end_time < t and
                (not self.matched_chain_pair or
                 chain is not self.matched_chain_pair[1])):
                # don't delete the active (matched) chain
                old_chains.append(chain)
        return old_chains

    def get_signature_chains_older_than(self, t):
        u"""
        Returns a list of signature chains older than the given time t

        All of the times will be associated with an intact chain.
        Furthermore, none of the times will be of a chain which a newer
        set may depend on.  For instance, if set A is a full set older
        than t, and set B is an incremental based on A which is newer
        than t, then the time of set A will not be returned.
        """
        assert self.values_set
        old_chains = []
        for chain in self.all_sig_chains:
            if (chain.end_time < t and
                (not self.matched_chain_pair or
                 chain is not self.matched_chain_pair[0])):
                # don't delete the active (matched) chain
                old_chains.append(chain)
        return old_chains

    def get_last_full_backup_time(self):
        u"""
        Return the time of the last full backup,
        or 0 if there is none.
        """
        return self.get_nth_last_full_backup_time(1)

    def get_nth_last_full_backup_time(self, n):
        u"""
        Return the time of the nth to last full backup,
        or 0 if there is none.
        """
        chain = self.get_nth_last_backup_chain(n)
        if chain is None:
            return 0
        else:
            return chain.get_first().time

    def get_last_backup_chain(self):
        u"""
        Return the last full backup of the collection,
        or None if there is no full backup chain.
        """
        return self.get_nth_last_backup_chain(1)

    def get_nth_last_backup_chain(self, n):
        u"""
        Return the nth-to-last full backup of the collection,
        or None if there is less than n backup chains.

        NOTE: n = 1 -> time of latest available chain (n = 0 is not
        a valid input). Thus the second-to-last is obtained with n=2
        rather than n=1.
        """
        assert self.values_set
        assert n > 0

        if len(self.all_backup_chains) < n:
            return None

        sorted = self.all_backup_chains[:]
        sorted.sort(key=lambda x: x.get_first().time)

        sorted.reverse()
        return sorted[n - 1]

    def get_older_than(self, t):
        u"""
        Returns a list of backup sets older than the given time t

        All of the times will be associated with an intact chain.
        Furthermore, none of the times will be of a set which a newer
        set may depend on.  For instance, if set A is a full set older
        than t, and set B is an incremental based on A which is newer
        than t, then the time of set A will not be returned.
        """
        old_sets = []
        for chain in self.get_chains_older_than(t):
            old_sets.extend(chain.get_all_sets())
        return self.sort_sets(old_sets)

    def get_older_than_required(self, t):
        u"""
        Returns list of old backup sets required by new sets

        This function is similar to the previous one, but it only
        returns the times of sets which are old but part of the chains
        where the newer end of the chain is newer than t.
        """
        assert self.values_set
        new_chains = [c for c in self.all_backup_chains if c.end_time >= t]
        result_sets = []
        for chain in new_chains:
            old_sets = [s for s in chain.get_all_sets() if s.get_time() < t]
            result_sets.extend(old_sets)
        return self.sort_sets(result_sets)

    def get_file_changed_record(self, filepath):
        u"""
        Returns time line of specified file changed
        """
        # quick fix to spaces in filepath
        modified_filepath = filepath
        if u" " in filepath:
            modified_filepath = u'"' + filepath.replace(u" ", r"\x20") + u'"'

        if not self.matched_chain_pair:
            return u""

        all_backup_set = self.matched_chain_pair[1].get_all_sets()
        specified_file_backup_set = []
        specified_file_backup_type = []

        for bs in all_backup_set:
            filelist = [fileinfo[1] for fileinfo in bs.get_files_changed()]
            if modified_filepath in filelist:
                specified_file_backup_set.append(bs)
                index = filelist.index(modified_filepath)
                specified_file_backup_type.append(bs.get_files_changed()[index][0])

        return FileChangedStatus(filepath, list(zip(specified_file_backup_type, specified_file_backup_set)))


class FileChangedStatus(object):
    def __init__(self, filepath, fileinfo_list):
        self.filepath = filepath
        self.fileinfo_list = fileinfo_list

    def __str__(self):
        set_schema = u"%20s   %30s  %20s"
        l = [u"-------------------------",
             _(u"File: %s") % (self.filepath),
             _(u"Total number of backup: %d") % len(self.fileinfo_list),
             set_schema % (_(u"Type of backup set:"), _(u"Time:"), _(u"Type of file change:"))]

        for s in self.fileinfo_list:
            backup_type = s[0]
            backup_set = s[1]
            if backup_set.time:
                type = _(u"Full")
            else:
                type = _(u"Incremental")
            l.append(set_schema % (type, dup_time.timetopretty(backup_set.get_time()), backup_type.title()))

        l.append(u"-------------------------")
        return u"\n".join(l)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Store global configuration information"""

import os
import sys
import socket

from duplicity import __version__


# The current version of duplicity
version = __version__

# Prefix for all files (appended before type-specific prefixes)
file_prefix = b""

# Prefix for manifest files only
file_prefix_manifest = b""

# Prefix for archive files only
file_prefix_archive = b""

# Prefix for sig files only
file_prefix_signature = b""

# The name of the current host, or None if it cannot be set
hostname = socket.getfqdn()

# The main local path.  For backing up the is the path to be backed
# up.  For restoring, this is the destination of the restored files.
local_path = None

# The symbolic name of the backup being operated upon.
backup_name = None

# For testing -- set current time
current_time = None

# Set to the Path of the archive directory (the directory which
# contains the signatures and manifests of the relevent backup
# collection), and for checkpoint state between volumes.
# NOTE: this gets expanded in duplicity.commandline
os.environ[u"XDG_CACHE_HOME"] = os.getenv(u"XDG_CACHE_HOME", os.path.expanduser(u"~/.cache"))
archive_dir = os.path.expandvars(u"$XDG_CACHE_HOME/duplicity")
archive_dir_path = None

# config dir for future use
os.environ[u"XDG_CONFIG_HOME"] = os.getenv(u"XDG_CONFIG_HOME", os.path.expanduser(u"~/.config"))
config_dir = os.path.expandvars(u"$XDG_CONFIG_HOME/duplicity")

# Restores will try to bring back the state as of the following time.
# If it is None, default to current time.
restore_time = None

# If set, restore only the subdirectory or file specified, not the
# whole root.
restore_dir = None

# The backend representing the remote side
backend = None

# Are errors fatal (set for retry decorator in backend.py)
# See example of use in multibackend.py _list()
# Do not use in normal cases!
are_errors_fatal = {
    u'delete': (True, None),
    u'get': (True, None),
    u'list': (True, None),
    u'move': (True, None),
    u'put': (True, None),
    u'query': (True, None),
}

# If set, the Select object which iterates paths in the local
# source directory.
select = None

gpg_binary = None

# Set to GPGProfile that will be used to compress/uncompress encrypted
# files.  Replaces encryption_keys, sign_key, and passphrase settings.
gpg_profile = None

# Options to pass to gpg
gpg_options = u''

# Maximum file blocksize
max_blocksize = 2048

# If true, filelists and directory statistics will be split on
# nulls instead of newlines.
null_separator = None

# number of retries on network operations
num_retries = 5

# True if Pydev debugger should be activated
pydevd = False

# Character used like the ":" in time strings like
# 2002-08-06T04:22:00-07:00.  The colon isn't good for filenames on
# windows machines.
time_separator = u":"

# Global lockfile used to manage concurrency
lockpath = u""
lockfile = None

# If this is true, only warn and don't raise fatal error when backup
# source directory doesn't match previous backup source directory.
allow_source_mismatch = None

# If set, abort if cannot do an incremental backup.  Otherwise if
# signatures not found, default to full.
incremental = None

# If set, print the statistics after every backup session
print_statistics = True

# If set, use short (< 30 char) filenames for all the remote files.
short_filenames = False

# If set, forces a full backup if the last full backup is older than
# the time specified
full_force_time = None

# Used to confirm certain destructive operations like deleting old files.
force = None

# If set, signifies time in seconds before which backup files should
# be deleted.
remove_time = None

# If set, signifies the number of backups chains to keep when performing
# a remove-all-but-n-full.
keep_chains = None

# If set, signifies that remove-all-but-n-full in progress
remove_all_but_n_full_mode = None

# If set, signifies that remove-all-inc-of-but-n-full in progress (variant of remove-all-but-n-full)
remove_all_inc_of_but_n_full_mode = None

# Don't actually do anything, but still report what would be done
dry_run = False

# If set to false, then do not encrypt files on remote system
encryption = True

# If set to false, then do not compress files on remote system
compression = True

# volume size. default 200M
volsize = 200 * 1024 * 1024

# Working directory for the tempfile module. Defaults to /tmp on most systems.
temproot = None

# network timeout value
timeout = 30

# FTP data connection type
ftp_connection = u'passive'

# Protocol for webdav
webdav_proto = u'http'

# Asynchronous put/get concurrency limit
# (default of 0 disables asynchronicity).
async_concurrency = 0

# Whether to use "new-style" subdomain addressing for S3 buckets. Such
# use is not backwards-compatible with upper-case buckets, or buckets
# that are otherwise not expressable in a valid hostname.
s3_use_new_style = False

# Whether to create European buckets (sorry, hard-coded to only
# support european for now).
s3_european_buckets = False

# File owner uid keeps number from tar file. Like same option in GNU tar.
numeric_owner = False

# Whether to use plain HTTP (without SSL) to send data to S3
# See <https://bugs.launchpad.net/duplicity/+bug/433970>.
s3_unencrypted_connection = False

# Whether to use S3 Reduced Redudancy Storage
s3_use_rrs = False

# Whether to use S3 Infrequent Access Storage
s3_use_ia = False

# Whether to use S3 Glacier Storage
s3_use_glacier = False

# Whether to use S3 Glacier Deep Archive Storage
s3_use_deep_archive = False

# Whether to use S3 One Zone Infrequent Access Storage
s3_use_onezone_ia = False

# True if we should use boto multiprocessing version
s3_use_multiprocessing = False

# Chunk size used for S3 multipart uploads.The number of parallel uploads to
# S3 be given by chunk size / volume size. Use this to maximize the use of
# your bandwidth. Defaults to 25MB
s3_multipart_chunk_size = 25 * 1024 * 1024

# Minimum chunk size accepted by S3
s3_multipart_minimum_chunk_size = 5 * 1024 * 1024

# Maximum number of processes to use while doing a multipart upload to S3
s3_multipart_max_procs = None

# Maximum time to wait for a part to finish when doig a multipart upload to S3
s3_multipart_max_timeout = None

# Use server side encryption in s3
s3_use_sse = False

# Use server side kms encryption in s3
s3_use_sse_kms = False
s3_kms_key_id = None
s3_kms_grant = None

# Which storage policy to use for Swift containers
swift_storage_policy = u""

# The largest size upload supported in a single put call for azure
azure_max_single_put_size = None

# The size of the blocks put to azure blob storage if bigger than azure_max_single_put_size
azure_max_block_size = None

# Maximum number of parallel connections to use when the blob size for azure exceeds 64MB
azure_max_connections = None

# Standard storage tier used for storring backup blobs (Hot|Cool|Archive).
azure_blob_tier = None

# Whether to use the full email address as the user name when
# logging into an imap server. If false just the user name
# part of the email address is used.
imap_full_address = False

# Name of the imap folder where we want to store backups.
# Can be changed with a command line argument.
imap_mailbox = u"INBOX"

# Sync partial metadata by default
metadata_sync_mode = u"partial"

# Whether the old filename format is in effect.
old_filenames = False

# Wheter to specify --use-agent in GnuPG options
use_agent = False

# ssh commands to use, used by ssh_pexpect (defaults to sftp, scp)
scp_command = None
sftp_command = None

# default to batch mode using public-key encryption
ssh_askpass = False

# user added ssh options
ssh_options = u""

# default cf backend is pyrax
cf_backend = u"pyrax"

# HTTPS ssl options (currently only webdav, lftp)
ssl_cacert_file = None
ssl_cacert_path = None
ssl_no_check_certificate = False

# user added rsync options
rsync_options = u""

# will be a Restart object if restarting
restart = None

# used in testing only - raises exception after volume
fail_on_volume = 0

# used in testing only - skips uploading a particular volume
skip_volume = 0

# ignore (some) errors during operations; supposed to make it more
# likely that you are able to restore data under problematic
# circumstances. the default should absolutely always be True unless
# you know what you are doing.
ignore_errors = False

# If we should be particularly aggressive when cleaning up
extra_clean = False

# Renames (--rename)
rename = {}

# enable data comparison on verify runs
compare_data = False

# When symlinks are encountered, the item they point to is copied rather than
# the symlink.
copy_links = False

# When selected, triggers a dry-run before a full or incremental to compute
# changes, then runs the real operation and keeps track of the real progress
progress = False

# Controls the upload progress messages refresh rate. Default: update each
# 3 seconds
progress_rate = 3

# Level of Redundancy in % for Par2 files
par2_redundancy = 10

# Verbatim par2 other options
par2_options = u""

# Whether to enable gio backend
use_gio = False

# If set, collect only the file status, not the whole root.
file_changed = None

# delay (in seconds) before next operation after failure
backend_retry_delay = 30

# default filesystem encoding
# In Python 2 it seems that sys.getfilesystemencoding() will normally return
# 'utf-8' or some other sane encoding, but will sometimes fail and return
# either 'ascii' or None.  Both are bogus, so default to 'utf-8' if it does.
fsencoding = sys.getfilesystemencoding()
fsencoding = fsencoding if fsencoding not in [u'ascii', u'ANSI_X3.4-1968', None] else u'utf-8'



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from past.builtins import cmp
from builtins import object
from future import standard_library
standard_library.install_aliases()

import sys
import unittest
import time
import types
from duplicity import dup_time
from . import UnitTestCase

# For type testing against both int and long types that works in python 2/3
if sys.version_info < (3,):
    integer_types = (int, int)
else:
    integer_types = (int,)


class TimeTest(object):
    def testConversion(self):
        u"""test timetostring and stringtotime"""
        dup_time.setcurtime()
        assert type(dup_time.curtime) in integer_types
        assert isinstance(dup_time.curtimestr, (str, u"".__class__))
        assert (dup_time.cmp(int(dup_time.curtime), dup_time.curtimestr) == 0 or
                dup_time.cmp(int(dup_time.curtime) + 1, dup_time.curtimestr) == 0)
        time.sleep(1.05)
        assert dup_time.cmp(time.time(), dup_time.curtime) == 1
        assert dup_time.cmp(dup_time.timetostring(time.time()), dup_time.curtimestr) == 1

    def testConversion_separator(self):
        u"""Same as testConversion, but change time Separator"""
        self.set_global(u'time_separator', u"_")
        self.testConversion()

    def testCmp(self):
        u"""Test time comparisons"""
        cmp = dup_time.cmp
        assert cmp(1, 2) == -1
        assert cmp(2, 2) == 0
        assert cmp(5, 1) == 1
        assert cmp(u"2001-09-01T21:49:04Z", u"2001-08-01T21:49:04Z") == 1
        assert cmp(u"2001-09-01T04:49:04+03:23", u"2001-09-01T21:49:04Z") == -1
        assert cmp(u"2001-09-01T12:00:00Z", u"2001-09-01T04:00:00-08:00") == 0
        assert cmp(u"2001-09-01T12:00:00-08:00", u"2001-09-01T12:00:00-07:00") == 1
        assert cmp(u"2001-09-01T11:00:00Z", u"20010901T120000Z") == -1
        assert cmp(u"2001-09-01T12:00:00Z", u"20010901T120000Z") == 0
        assert cmp(u"2001-09-01T13:00:00Z", u"20010901T120000Z") == 1

    def testCmp_separator(self):
        u"""Like testCmp but with new separator"""
        self.set_global(u'time_separator', u"_")
        cmp = dup_time.cmp
        assert cmp(1, 2) == -1
        assert cmp(2, 2) == 0
        assert cmp(5, 1) == 1
        assert cmp(u"2001-09-01T21_49_04Z", u"2001-08-01T21_49_04Z") == 1
        assert cmp(u"2001-09-01T04_49_04+03_23", u"2001-09-01T21_49_04Z") == -1
        assert cmp(u"2001-09-01T12_00_00Z", u"2001-09-01T04_00_00-08_00") == 0
        assert cmp(u"2001-09-01T12_00_00-08_00", u"2001-09-01T12_00_00-07_00") == 1

    def testStringtotime(self):
        u"""Test converting string to time"""
        timesec = int(time.time())
        assert timesec == int(dup_time.stringtotime(dup_time.timetostring(timesec)))
        assert not dup_time.stringtotime(u"2001-18-83T03:03:03Z")
        assert not dup_time.stringtotime(u"2001-01-23L03:03:03L")
        assert not dup_time.stringtotime(u"2001_01_23T03:03:03Z")

    def testIntervals(self):
        u"""Test converting strings to intervals"""
        i2s = dup_time.intstringtoseconds
        for s in [u"32", u"", u"d", u"231I", u"MM", u"s", u"-2h"]:
            try:
                i2s(s)
            except dup_time.TimeException:
                pass
            else:
                assert 0, s
        assert i2s(u"7D") == 7 * 86400
        assert i2s(u"232s") == 232
        assert i2s(u"2M") == 2 * 30 * 86400
        assert i2s(u"400m") == 400 * 60
        assert i2s(u"1Y") == 365 * 86400
        assert i2s(u"30h") == 30 * 60 * 60
        assert i2s(u"3W") == 3 * 7 * 86400

    def testIntervalsComposite(self):
        u"""Like above, but allow composite intervals"""
        i2s = dup_time.intstringtoseconds
        assert i2s(u"7D2h") == 7 * 86400 + 2 * 3600
        assert i2s(u"2Y3s") == 2 * 365 * 86400 + 3
        assert i2s(u"1M2W4D2h5m20s") == (30 * 86400 + 2 * 7 * 86400 + 4 * 86400 +
                                        2 * 3600 + 5 * 60 + 20)

    def testPrettyIntervals(self):
        u"""Test printable interval conversion"""
        assert dup_time.inttopretty(3600) == u"1 hour"
        assert dup_time.inttopretty(7220) == u"2 hours 20 seconds"
        assert dup_time.inttopretty(0) == u"0 seconds"
        assert dup_time.inttopretty(353) == u"5 minutes 53 seconds"
        assert dup_time.inttopretty(3661) == u"1 hour 1 minute 1 second"
        assert dup_time.inttopretty(353.234234) == u"5 minutes 53.23 seconds"

    def testGenericString(self):
        u"""Test genstrtotime, conversion of arbitrary string to time"""
        g2t = dup_time.genstrtotime
        assert g2t(u'now', 1000) == 1000
        assert g2t(u'2h3s', 10000) == 10000 - 2 * 3600 - 3
        assert g2t(u'2001-09-01T21:49:04Z') == \
            dup_time.stringtotime(u'2001-09-01T21:49:04Z')
        assert g2t(u'2002-04-26T04:22:01') == \
            dup_time.stringtotime(u'2002-04-26T04:22:01' + dup_time.gettzd(0))
        t = dup_time.stringtotime(u'2001-05-12T00:00:00' + dup_time.gettzd(0))
        assert g2t(u'2001-05-12') == t
        assert g2t(u'2001/05/12') == t
        assert g2t(u'5/12/2001') == t
        assert g2t(u'123456') == 123456

    def testGenericStringErrors(self):
        u"""Test genstrtotime on some bad strings"""
        g2t = dup_time.genstrtotime
        self.assertRaises(dup_time.TimeException, g2t, u"hello")
        self.assertRaises(dup_time.TimeException, g2t, u"")
        self.assertRaises(dup_time.TimeException, g2t, u"3q")

    def testConvertion(self):
        t = int(time.time())
        assert dup_time.stringtotime(dup_time.timetostring(t)) == t


class TimeTest1(TimeTest, UnitTestCase):

    def setUp(self):
        super(TimeTest1, self).setUp()
        self.set_global(u'old_filenames', False)


class TimeTest2(TimeTest, UnitTestCase):

    def setUp(self):
        super(TimeTest2, self).setUp()
        self.set_global(u'old_filenames', True)

if __name__ == u'__main__':
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""
Provides a common interface to all backends and certain sevices
intended to be used by the backends themselves.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object

import errno
import os
import sys
import time
import re
import getpass
import re
import types
import urllib.request  # pylint: disable=import-error
import urllib.parse  # pylint: disable=import-error
import urllib.error  # pylint: disable=import-error

from duplicity import dup_temp
from duplicity import file_naming
from duplicity import globals
from duplicity import log
from duplicity import path
from duplicity import progress
from duplicity import util

from duplicity.util import exception_traceback

from duplicity.errors import BackendException
from duplicity.errors import FatalBackendException
from duplicity.errors import TemporaryLoadException
from duplicity.errors import ConflictingScheme
from duplicity.errors import InvalidBackendURL
from duplicity.errors import UnsupportedBackendScheme

import duplicity.backends

_backends = {}
_backend_prefixes = {}
_last_exception = None

# These URL schemes have a backend with a notion of an RFC "network location".
# The 'file' and 's3+http' schemes should not be in this list.
# 'http' and 'https' are not actually used for duplicity backend urls, but are needed
# in order to properly support urls returned from some webdav servers. adding them here
# is a hack. we should instead not stomp on the url parsing module to begin with.
#
# This looks similar to urlparse's 'uses_netloc' list, but urlparse doesn't use
# that list for parsing, only creating urls.  And doesn't include our custom
# schemes anyway.  So we keep our own here for our own use.
#
# NOTE: this is filled by the respective backends during registering
uses_netloc = []


def import_backends():
    u"""
    Import files in the duplicity/backends directory where
    the filename ends in 'backend.py' and ignore the rest.

    @rtype: void
    @return: void
    """
    path = duplicity.backends.__path__[0]
    assert path.endswith(u"duplicity/backends"), duplicity.backends.__path__

    files = os.listdir(path)
    files.sort()
    for fn in files:
        if fn.endswith(u"backend.py"):
            fn = fn[:-3]
            imp = u"duplicity.backends.%s" % (fn,)
            try:
                __import__(imp)
                res = u"Succeeded"
            except Exception:
                res = u"Failed: " + str(sys.exc_info()[1])
            log.Log(_(u"Import of %s %s") % (imp, res), log.INFO)
        else:
            continue


def register_backend(scheme, backend_factory):
    u"""
    Register a given backend factory responsible for URL:s with the
    given scheme.

    The backend must be a callable which, when called with a URL as
    the single parameter, returns an object implementing the backend
    protocol (i.e., a subclass of Backend).

    Typically the callable will be the Backend subclass itself.

    This function is not thread-safe and is intended to be called
    during module importation or start-up.
    """
    global _backends

    assert callable(backend_factory), u"backend factory must be callable"

    if scheme in _backends:
        raise ConflictingScheme(u"the scheme %s already has a backend "
                                u"associated with it"
                                u"" % (scheme,))

    _backends[scheme] = backend_factory


def register_backend_prefix(scheme, backend_factory):
    u"""
    Register a given backend factory responsible for URL:s with the
    given scheme prefix.

    The backend must be a callable which, when called with a URL as
    the single parameter, returns an object implementing the backend
    protocol (i.e., a subclass of Backend).

    Typically the callable will be the Backend subclass itself.

    This function is not thread-safe and is intended to be called
    during module importation or start-up.
    """
    global _backend_prefixes

    assert callable(backend_factory), u"backend factory must be callable"

    if scheme in _backend_prefixes:
        raise ConflictingScheme(u"the prefix %s already has a backend "
                                u"associated with it"
                                u"" % (scheme,))

    _backend_prefixes[scheme] = backend_factory


def strip_prefix(url_string, prefix_scheme):
    u"""
    strip the prefix from a string e.g. par2+ftp://... -> ftp://...
    """
    return re.sub(r'(?i)^' + re.escape(prefix_scheme) + r'\+', r'', url_string)


def is_backend_url(url_string):
    u"""
    @return Whether the given string looks like a backend URL.
    """
    pu = ParsedUrl(url_string)

    # Be verbose to actually return True/False rather than string.
    if pu.scheme:
        return True
    else:
        return False


def get_backend_object(url_string):
    u"""
    Find the right backend class instance for the given URL, or return None
    if the given string looks like a local path rather than a URL.

    Raise InvalidBackendURL if the URL is not a valid URL.
    """
    if not is_backend_url(url_string):
        return None

    global _backends, _backend_prefixes

    pu = ParsedUrl(url_string)
    assert pu.scheme, u"should be a backend url according to is_backend_url"

    factory = None

    for prefix in _backend_prefixes:
        if url_string.startswith(prefix + u'+'):
            factory = _backend_prefixes[prefix]
            pu = ParsedUrl(strip_prefix(url_string, prefix))
            break

    if factory is None:
        if pu.scheme not in _backends:
            raise UnsupportedBackendScheme(url_string)
        else:
            factory = _backends[pu.scheme]

    try:
        return factory(pu)
    except ImportError:
        raise BackendException(_(u"Could not initialize backend: %s") % str(sys.exc_info()[1]))


def get_backend(url_string):
    u"""
    Instantiate a backend suitable for the given URL, or return None
    if the given string looks like a local path rather than a URL.

    Raise InvalidBackendURL if the URL is not a valid URL.
    """
    if globals.use_gio:
        url_string = u'gio+' + url_string
    obj = get_backend_object(url_string)
    if obj:
        obj = BackendWrapper(obj)
    return obj


class ParsedUrl(object):
    u"""
    Parse the given URL as a duplicity backend URL.

    Returns the data of a parsed URL with the same names as that of
    the standard urlparse.urlparse() except that all values have been
    resolved rather than deferred.  There are no get_* members.  This
    makes sure that the URL parsing errors are detected early.

    Raise InvalidBackendURL on invalid URL's
    """
    def __init__(self, url_string):
        self.url_string = url_string

        # Python < 2.6.5 still examine urlparse.uses_netlock when parsing urls,
        # so stuff our custom list in there before we parse.
        urllib.parse.uses_netloc = uses_netloc

        # While useful in some cases, the fact is that the urlparser makes
        # all the properties in the URL deferred or lazy.  This means that
        # problems don't get detected till called.  We'll try to trap those
        # problems here, so they will be caught early.

        try:
            pu = urllib.parse.urlparse(url_string)
        except Exception:
            raise InvalidBackendURL(u"Syntax error in: %s" % url_string)

        try:
            self.scheme = pu.scheme
        except Exception:
            raise InvalidBackendURL(u"Syntax error (scheme) in: %s" % url_string)

        try:
            self.netloc = pu.netloc
        except Exception:
            raise InvalidBackendURL(u"Syntax error (netloc) in: %s" % url_string)

        try:
            self.path = pu.path
            if self.path:
                self.path = urllib.parse.unquote(self.path)
        except Exception:
            raise InvalidBackendURL(u"Syntax error (path) in: %s" % url_string)

        try:
            self.username = pu.username
        except Exception:
            raise InvalidBackendURL(u"Syntax error (username) in: %s" % url_string)
        if self.username:
            self.username = urllib.parse.unquote(pu.username)
        else:
            self.username = None

        try:
            self.password = pu.password
        except Exception:
            raise InvalidBackendURL(u"Syntax error (password) in: %s" % url_string)
        if self.password:
            self.password = urllib.parse.unquote(self.password)
        else:
            self.password = None

        try:
            self.hostname = pu.hostname
        except Exception:
            raise InvalidBackendURL(u"Syntax error (hostname) in: %s" % url_string)

        # init to None, overwrite with actual value on success
        self.port = None
        try:
            self.port = pu.port
        except Exception:  # not raised in python2.7, just returns None
            # TODO: remove after dropping python 2.7 support
            if self.scheme in [u'rclone']:
                pass
            # old style rsync://host::[/]dest, are still valid, though they contain no port
            elif not (self.scheme in [u'rsync'] and re.search(u'::[^:]*$', self.url_string)):
                raise InvalidBackendURL(u"Syntax error (port) in: %s A%s B%s C%s" %
                                        (url_string, (self.scheme in [u'rsync']),
                                         re.search(u'::[^:]+$', self.netloc), self.netloc))

        # Our URL system uses two slashes more than urlparse's does when using
        # non-netloc URLs.  And we want to make sure that if urlparse assuming
        # a netloc where we don't want one, that we correct it.
        if self.scheme not in uses_netloc:
            if self.netloc:
                self.path = u'//' + self.netloc + self.path
                self.netloc = u''
                self.hostname = None
            elif not self.path.startswith(u'//') and self.path.startswith(u'/'):
                self.path = u'//' + self.path

        # This happens for implicit local paths.
        if not self.scheme:
            return

        # Our backends do not handle implicit hosts.
        if self.scheme in uses_netloc and not self.hostname:
            raise InvalidBackendURL(u"Missing hostname in a backend URL which "
                                    u"requires an explicit hostname: %s"
                                    u"" % (url_string))

        # Our backends do not handle implicit relative paths.
        if self.scheme not in uses_netloc and not self.path.startswith(u'//'):
            raise InvalidBackendURL(u"missing // - relative paths not supported "
                                    u"for scheme %s: %s"
                                    u"" % (self.scheme, url_string))

    def geturl(self):
        return self.url_string


def strip_auth_from_url(parsed_url):
    u"""Return a URL from a urlparse object without a username or password."""

    clean_url = re.sub(u'^([^:/]+://)(.*@)?(.*)', r'\1\3', parsed_url.geturl())
    return clean_url


def _get_code_from_exception(backend, operation, e):
    if isinstance(e, BackendException) and e.code != log.ErrorCode.backend_error:
        return e.code
    elif hasattr(backend, u'_error_code'):
        return backend._error_code(operation, e) or log.ErrorCode.backend_error
    elif hasattr(e, u'errno'):
        # A few backends return such errors (local, paramiko, etc)
        if e.errno == errno.EACCES:
            return log.ErrorCode.backend_permission_denied
        elif e.errno == errno.ENOENT:
            return log.ErrorCode.backend_not_found
        elif e.errno == errno.ENOSPC:
            return log.ErrorCode.backend_no_space
    return log.ErrorCode.backend_error


def retry(operation, fatal=True):
    # Decorators with arguments introduce a new level of indirection.  So we
    # have to return a decorator function (which itself returns a function!)
    def outer_retry(fn):
        def inner_retry(self, *args):
            global _last_exception
            errors_fatal, errors_default = globals.are_errors_fatal.get(operation, (True, None))
            for n in range(1, globals.num_retries + 1):
                try:
                    return fn(self, *args)
                except FatalBackendException as e:
                    _last_exception = e
                    if not errors_fatal:
                        # backend wants to report and ignore errors
                        return errors_default
                    else:
                        # die on fatal errors
                        raise e
                except Exception as e:
                    _last_exception = e
                    if not errors_fatal:
                        # backend wants to report and ignore errors
                        return errors_default
                    else:
                        # retry on anything else
                        log.Debug(_(u"Backtrace of previous error: %s")
                                  % exception_traceback())
                        at_end = n == globals.num_retries
                        code = _get_code_from_exception(self.backend, operation, e)
                        if code == log.ErrorCode.backend_not_found:
                            # If we tried to do something, but the file just isn't there,
                            # no need to retry.
                            at_end = True
                        if at_end and fatal:
                            def make_filename(f):
                                if isinstance(f, path.ROPath):
                                    return util.escape(f.uc_name)
                                else:
                                    return util.escape(f)
                            extra = u' '.join([operation] + [make_filename(x) for x in args
                                                             if (x and isinstance(x, str))])
                            log.FatalError(_(u"Giving up after %s attempts. %s: %s")
                                           % (n, e.__class__.__name__,
                                              util.uexc(e)), code=code, extra=extra)
                        else:
                            log.Warn(_(u"Attempt %s failed. %s: %s")
                                     % (n, e.__class__.__name__, util.uexc(e)))
                        if not at_end:
                            if isinstance(e, TemporaryLoadException):
                                time.sleep(3 * globals.backend_retry_delay)  # wait longer before trying again
                            else:
                                time.sleep(globals.backend_retry_delay)  # wait a bit before trying again
                            if hasattr(self.backend, u'_retry_cleanup'):
                                self.backend._retry_cleanup()

        return inner_retry
    return outer_retry


class Backend(object):
    u"""
    See README in backends directory for information on how to write a backend.
    """
    def __init__(self, parsed_url):
        self.parsed_url = parsed_url

    u""" use getpass by default, inherited backends may overwrite this behaviour """
    use_getpass = True

    def get_password(self):
        u"""
        Return a password for authentication purposes. The password
        will be obtained from the backend URL, the environment, by
        asking the user, or by some other method. When applicable, the
        result will be cached for future invocations.
        """
        if self.parsed_url.password:
            return self.parsed_url.password

        try:
            password = os.environ[u'FTP_PASSWORD']
        except KeyError:
            if self.use_getpass:
                password = getpass.getpass(u"Password for '%s@%s': " %
                                           (self.parsed_url.username, self.parsed_url.hostname))
                os.environ[u'FTP_PASSWORD'] = password
            else:
                password = None
        return password

    def munge_password(self, commandline):
        u"""
        Remove password from commandline by substituting the password
        found in the URL, if any, with a generic place-holder.

        This is intended for display purposes only, and it is not
        guaranteed that the results are correct (i.e., more than just
        the ':password@' may be substituted.
        """
        if self.parsed_url.password:
            return re.sub(r'(:([^\s:/@]+)@([^\s@]+))', r':*****@\3', commandline)
        else:
            return commandline

    def __subprocess_popen(self, args):
        u"""
        For internal use.
        Execute the given command line, interpreted as a shell command.
        Returns int Exitcode, string StdOut, string StdErr
        """
        from subprocess import Popen, PIPE

        args[0] = util.which(args[0])
        p = Popen(args, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        stdout, stderr = p.communicate()

        return p.returncode, stdout, stderr

    u""" a dictionary for breaking exceptions, syntax is
        { 'command' : [ code1, code2 ], ... } see ftpbackend for an example """
    popen_breaks = {}

    def subprocess_popen(self, commandline):
        u"""
        Execute the given command line with error check.
        Returns int Exitcode, string StdOut, string StdErr

        Raise a BackendException on failure.
        """
        import shlex

        if isinstance(commandline, (list, tuple)):
            logstr = u' '.join(commandline)
            args = commandline
        else:
            logstr = commandline
            args = shlex.split(commandline)

        logstr = self.munge_password(logstr)
        log.Info(_(u"Reading results of '%s'") % logstr)

        result, stdout, stderr = self.__subprocess_popen(args)
        if result != 0:
            try:
                ignores = self.popen_breaks[args[0]]
                ignores.index(result)
                u""" ignore a predefined set of error codes """
                return 0, u'', u''
            except (KeyError, ValueError):
                raise BackendException(u"Error running '%s': returned %d, with output:\n%s" %
                                       (logstr, result, stdout + u'\n' + stderr + u'\n'))
        return result, stdout, stderr


class BackendWrapper(object):
    u"""
    Represents a generic duplicity backend, capable of storing and
    retrieving files.
    """

    def __init__(self, backend):
        self.backend = backend

    def __do_put(self, source_path, remote_filename):
        if hasattr(self.backend, u'_put'):
            log.Info(_(u"Writing %s") % util.fsdecode(remote_filename))
            self.backend._put(source_path, remote_filename)
        else:
            raise NotImplementedError()

    @retry(u'put', fatal=True)
    def put(self, source_path, remote_filename=None):
        u"""
        Transfer source_path (Path object) to remote_filename (string)

        If remote_filename is None, get the filename from the last
        path component of pathname.
        """
        if not remote_filename:
            remote_filename = source_path.get_filename()
        self.__do_put(source_path, remote_filename)

    @retry(u'move', fatal=True)
    def move(self, source_path, remote_filename=None):
        u"""
        Move source_path (Path object) to remote_filename (string)

        Same as put(), but unlinks source_path in the process.  This allows the
        local backend to do this more efficiently using rename.
        """
        if not remote_filename:
            remote_filename = source_path.get_filename()
        if hasattr(self.backend, u'_move'):
            if self.backend._move(source_path, remote_filename) is not False:
                source_path.setdata()
                return
        self.__do_put(source_path, remote_filename)
        source_path.delete()

    @retry(u'get', fatal=True)
    def get(self, remote_filename, local_path):
        u"""Retrieve remote_filename and place in local_path"""
        if hasattr(self.backend, u'_get'):
            self.backend._get(remote_filename, local_path)
            local_path.setdata()
            if not local_path.exists():
                raise BackendException(_(u"File %s not found locally after get "
                                         u"from backend") % local_path.uc_name)
        else:
            raise NotImplementedError()

    @retry(u'list', fatal=True)
    def list(self):
        u"""
        Return list of filenames (byte strings) present in backend
        """
        def tobytes(filename):
            u"Convert a (maybe unicode) filename to bytes"
            if isinstance(filename, str):
                # There shouldn't be any encoding errors for files we care
                # about, since duplicity filenames are ascii.  But user files
                # may be in the same directory.  So just replace characters.
                return util.fsencode(filename)
            else:
                return filename

        if hasattr(self.backend, u'_list'):
            # Make sure that duplicity internals only ever see byte strings
            # for filenames, no matter what the backend thinks it is talking.
            return [tobytes(x) for x in self.backend._list()]
        else:
            raise NotImplementedError()

    def delete(self, filename_list):
        u"""
        Delete each filename in filename_list, in order if possible.
        """
        assert not isinstance(filename_list, bytes)
        if hasattr(self.backend, u'_delete_list'):
            self._do_delete_list(filename_list)
        elif hasattr(self.backend, u'_delete'):
            for filename in filename_list:
                self._do_delete(filename)
        else:
            raise NotImplementedError()

    @retry(u'delete', fatal=False)
    def _do_delete_list(self, filename_list):
        while filename_list:
            sublist = filename_list[:100]
            self.backend._delete_list(sublist)
            filename_list = filename_list[100:]

    @retry(u'delete', fatal=False)
    def _do_delete(self, filename):
        self.backend._delete(filename)

    # Should never cause FatalError.
    # Returns a dictionary of dictionaries.  The outer dictionary maps
    # filenames to metadata dictionaries.  Supported metadata are:
    #
    # 'size': if >= 0, size of file
    #         if -1, file is not found
    #         if None, error querying file
    #
    # Returned dictionary is guaranteed to contain a metadata dictionary for
    # each filename, and all metadata are guaranteed to be present.
    def query_info(self, filename_list):
        u"""
        Return metadata about each filename in filename_list
        """
        info = {}
        if hasattr(self.backend, u'_query_list'):
            info = self._do_query_list(filename_list)
            if info is None:
                info = {}
        elif hasattr(self.backend, u'_query'):
            for filename in filename_list:
                info[filename] = self._do_query(filename)

        # Fill out any missing entries (may happen if backend has no support
        # or its query_list support is lazy)
        for filename in filename_list:
            if filename not in info or info[filename] is None:
                info[filename] = {}
            for metadata in [u'size']:
                info[filename].setdefault(metadata, None)

        return info

    @retry(u'query', fatal=False)
    def _do_query_list(self, filename_list):
        info = self.backend._query_list(filename_list)
        if info is None:
            info = {}
        return info

    @retry(u'query', fatal=False)
    def _do_query(self, filename):
        try:
            return self.backend._query(filename)
        except Exception as e:
            code = _get_code_from_exception(self.backend, u'query', e)
            if code == log.ErrorCode.backend_not_found:
                return {u'size': -1}
            else:
                raise e

    def close(self):
        u"""
        Close the backend, releasing any resources held and
        invalidating any file objects obtained from the backend.
        """
        if hasattr(self.backend, u'_close'):
            self.backend._close()

    def get_fileobj_read(self, filename, parseresults=None):
        u"""
        Return fileobject opened for reading of filename on backend

        The file will be downloaded first into a temp file.  When the
        returned fileobj is closed, the temp file will be deleted.
        """
        if not parseresults:
            parseresults = file_naming.parse(filename)
            assert parseresults, u"Filename not correctly parsed"
        tdp = dup_temp.new_tempduppath(parseresults)
        self.get(filename, tdp)
        tdp.setdata()
        return tdp.filtered_open_with_delete(u"rb")

    def get_data(self, filename, parseresults=None):
        u"""
        Retrieve a file from backend, process it, return contents.
        """
        fin = self.get_fileobj_read(filename, parseresults)
        buf = fin.read()
        assert not fin.close()
        return buf



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import mock
import unittest

import duplicity.backend
import duplicity.backends  # @UnusedImport
from duplicity.errors import *  # @UnusedWildImport
from duplicity import globals
from duplicity import path
from . import UnitTestCase


class ParsedUrlTest(UnitTestCase):
    u"""Test the ParsedUrl class"""
    def test_basic(self):
        u"""Test various url strings"""
        pu = duplicity.backend.ParsedUrl(u"scp://ben@foo.bar:1234/a/b")
        assert pu.scheme == u"scp", pu.scheme
        assert pu.netloc == u"ben@foo.bar:1234", pu.netloc
        assert pu.path == u"/a/b", pu.path
        assert pu.username == u"ben", pu.username
        assert pu.port == 1234, pu.port
        assert pu.hostname == u"foo.bar", pu.hostname

        pu = duplicity.backend.ParsedUrl(u"ftp://foo.bar:1234/")
        assert pu.scheme == u"ftp", pu.scheme
        assert pu.netloc == u"foo.bar:1234", pu.netloc
        assert pu.path == u"/", pu.path
        assert pu.username is None, pu.username
        assert pu.port == 1234, pu.port
        assert pu.hostname == u"foo.bar", pu.hostname

        pu = duplicity.backend.ParsedUrl(u"file:///home")
        assert pu.scheme == u"file", pu.scheme
        assert pu.netloc == u"", pu.netloc
        assert pu.path == u"///home", pu.path
        assert pu.username is None, pu.username
        assert pu.port is None, pu.port

        pu = duplicity.backend.ParsedUrl(u"file://home")
        assert pu.scheme == u"file", pu.scheme
        assert pu.netloc == u"", pu.netloc
        assert pu.path == u"//home", pu.path
        assert pu.username is None, pu.username
        assert pu.port is None, pu.port

        pu = duplicity.backend.ParsedUrl(u"ftp://foo@bar:pass@example.com:123/home")
        assert pu.scheme == u"ftp", pu.scheme
        assert pu.netloc == u"foo@bar:pass@example.com:123", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar", pu.username
        assert pu.password == u"pass", pu.password
        assert pu.port == 123, pu.port

        pu = duplicity.backend.ParsedUrl(u"ftp://foo%40bar:pass@example.com:123/home")
        assert pu.scheme == u"ftp", pu.scheme
        assert pu.netloc == u"foo%40bar:pass@example.com:123", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar", pu.username
        assert pu.password == u"pass", pu.password
        assert pu.port == 123, pu.port

        pu = duplicity.backend.ParsedUrl(u"imap://foo@bar:pass@example.com:123/home")
        assert pu.scheme == u"imap", pu.scheme
        assert pu.netloc == u"foo@bar:pass@example.com:123", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar", pu.username
        assert pu.password == u"pass", pu.password
        assert pu.port == 123, pu.port

        pu = duplicity.backend.ParsedUrl(u"imap://foo@bar@example.com:123/home")
        assert pu.scheme == u"imap", pu.scheme
        assert pu.netloc == u"foo@bar@example.com:123", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar", pu.username
        assert pu.password is None, pu.password
        assert pu.port == 123, pu.port

        pu = duplicity.backend.ParsedUrl(u"imap://foo@bar@example.com/home")
        assert pu.scheme == u"imap", pu.scheme
        assert pu.netloc == u"foo@bar@example.com", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar", pu.username
        assert pu.password is None, pu.password
        assert pu.port is None, pu.port

        pu = duplicity.backend.ParsedUrl(u"imap://foo@bar.com@example.com/home")
        assert pu.scheme == u"imap", pu.scheme
        assert pu.netloc == u"foo@bar.com@example.com", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar.com", pu.username
        assert pu.password is None, pu.password
        assert pu.port is None, pu.port

        pu = duplicity.backend.ParsedUrl(u"imap://foo%40bar.com@example.com/home")
        assert pu.scheme == u"imap", pu.scheme
        assert pu.netloc == u"foo%40bar.com@example.com", pu.netloc
        assert pu.hostname == u"example.com", pu.hostname
        assert pu.path == u"/home", pu.path
        assert pu.username == u"foo@bar.com", pu.username
        assert pu.password is None, pu.password
        assert pu.port is None, pu.port

    def test_errors(self):
        u"""Test various url errors"""
        self.assertRaises(InvalidBackendURL, duplicity.backend.ParsedUrl,
                          u"file:path")  # no relative paths for non-netloc schemes
        self.assertRaises(UnsupportedBackendScheme, duplicity.backend.get_backend,
                          u"ssh://foo@bar:pass@example.com/home")


class BackendWrapperTest(UnitTestCase):

    def setUp(self):
        super(BackendWrapperTest, self).setUp()
        self.mock = mock.MagicMock()
        self.backend = duplicity.backend.BackendWrapper(self.mock)
        self.local = mock.MagicMock()
        self.remote = u'remote'

    @mock.patch(u'sys.exit')
    def test_default_error_exit(self, exit_mock):
        self.set_global(u'num_retries', 1)
        try:
            del self.mock._error_code
        except:
            # Old versions of mock don't let you mark non-present attributes
            # like this.
            return  # can't use self.skip() since that needs py27
        self.mock._put.side_effect = Exception
        self.backend.put(self.local, self.remote)
        exit_mock.assert_called_once_with(50)

    @mock.patch(u'sys.exit')
    def test_translates_code(self, exit_mock):
        self.set_global(u'num_retries', 1)
        self.mock._error_code.return_value = 12345
        self.mock._put.side_effect = Exception
        self.backend.put(self.local, self.remote)
        exit_mock.assert_called_once_with(12345)

    @mock.patch(u'sys.exit')
    def test_uses_exception_code(self, exit_mock):
        self.set_global(u'num_retries', 1)
        self.mock._error_code.return_value = 12345
        self.mock._put.side_effect = BackendException(u'error', code=54321)
        self.backend.put(self.local, self.remote)
        exit_mock.assert_called_once_with(54321)

    @mock.patch(u'sys.exit')
    @mock.patch(u'time.sleep')  # so no waiting
    def test_cleans_up(self, exit_mock, time_mock):
        self.set_global(u'num_retries', 2)
        self.mock._retry_cleanup.return_value = None
        self.mock._put.side_effect = Exception
        self.backend.put(self.local, self.remote)
        self.mock._retry_cleanup.assert_called_once_with()

    def test_prefer_lists(self):
        self.mock._delete.return_value = None
        self.mock._delete_list.return_value = None
        self.backend.delete([self.remote])
        self.assertEqual(self.mock._delete.call_count, 0)
        self.assertEqual(self.mock._delete_list.call_count, 1)
        try:
            del self.mock._delete_list
        except:
            return
        self.backend.delete([self.remote])
        self.assertEqual(self.mock._delete.call_count, 1)

        self.mock._query.return_value = None
        self.mock._query_list.return_value = None
        self.backend.query_info([self.remote])
        self.assertEqual(self.mock._query.call_count, 0)
        self.assertEqual(self.mock._query_list.call_count, 1)
        try:
            del self.mock._query_list
        except:
            return
        self.backend.query_info([self.remote])
        self.assertEqual(self.mock._query.call_count, 1)

    @mock.patch(u'sys.exit')
    @mock.patch(u'time.sleep')  # so no waiting
    def test_retries(self, exit_mock, time_mock):
        self.set_global(u'num_retries', 2)

        self.mock._get.side_effect = Exception
        self.backend.get(self.remote, self.local)
        self.assertEqual(self.mock._get.call_count, globals.num_retries)

        self.mock._put.side_effect = Exception
        self.backend.put(self.local, self.remote)
        self.assertEqual(self.mock._put.call_count, globals.num_retries)

        self.mock._list.side_effect = Exception
        self.backend.list()
        self.assertEqual(self.mock._list.call_count, globals.num_retries)

        self.mock._delete_list.side_effect = Exception
        self.backend.delete([self.remote])
        self.assertEqual(self.mock._delete_list.call_count, globals.num_retries)

        self.mock._query_list.side_effect = Exception
        self.backend.query_info([self.remote])
        self.assertEqual(self.mock._query_list.call_count, globals.num_retries)

        try:
            del self.mock._delete_list
        except:
            return
        self.mock._delete.side_effect = Exception
        self.backend.delete([self.remote])
        self.assertEqual(self.mock._delete.call_count, globals.num_retries)

        try:
            del self.mock._query_list
        except:
            return
        self.mock._query.side_effect = Exception
        self.backend.query_info([self.remote])
        self.assertEqual(self.mock._query.call_count, globals.num_retries)

        self.mock._move.side_effect = Exception
        self.backend.move(self.local, self.remote)
        self.assertEqual(self.mock._move.call_count, globals.num_retries)

    def test_move(self):
        self.mock._move.return_value = True
        self.backend.move(self.local, self.remote)
        self.mock._move.assert_called_once_with(self.local, self.remote)
        self.assertEqual(self.mock._put.call_count, 0)

    def test_move_fallback_false(self):
        self.mock._move.return_value = False
        self.backend.move(self.local, self.remote)
        self.mock._move.assert_called_once_with(self.local, self.remote)
        self.mock._put.assert_called_once_with(self.local, self.remote)
        self.local.delete.assert_called_once_with()

    def test_move_fallback_undefined(self):
        try:
            del self.mock._move
        except:
            return
        self.backend.move(self.local, self.remote)
        self.mock._put.assert_called_once_with(self.local, self.remote)
        self.local.delete.assert_called_once_with()

    def test_close(self):
        self.mock._close.return_value = None
        self.backend.close()
        self.mock._close.assert_called_once_with()


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016 Stefan Breunig <stefan-duplicity@breunig.xyz>
# Based on the backend onedrivebackend.py
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from __future__ import division
from builtins import input
import os.path
import json
import sys
import time
import re
from io import DEFAULT_BUFFER_SIZE

import duplicity.backend
from duplicity.errors import BackendException
from duplicity import globals
from duplicity import log


class ADBackend(duplicity.backend.Backend):
    u"""
    Backend for Amazon Drive. It communicates directly with Amazon Drive using
    their RESTful API and does not rely on externally setup software (like
    acd_cli).
    """

    OAUTH_TOKEN_PATH = os.path.expanduser(u'~/.duplicity_ad_oauthtoken.json')

    OAUTH_AUTHORIZE_URL = u'https://www.amazon.com/ap/oa'
    OAUTH_TOKEN_URL = u'https://api.amazon.com/auth/o2/token'
    # NOTE: Amazon requires https, which is why I am using my domain/setup
    # instead of Duplicity's. Mail me at stefan-duplicity@breunig.xyz once it is
    # available through https and I will whitelist the new URL.
    OAUTH_REDIRECT_URL = u'https://breunig.xyz/duplicity/copy.html'
    OAUTH_SCOPE = [u'clouddrive:read_other', u'clouddrive:write']

    CLIENT_ID = u'amzn1.application-oa2-client.791c9c2d78444e85a32eb66f92eb6bcc'
    CLIENT_SECRET = u'5b322c6a37b25f16d848a6a556eddcc30314fc46ae65c87068ff1bc4588d715b'

    MULTIPART_BOUNDARY = u'DuplicityFormBoundaryd66364f7f8924f7e9d478e19cf4b871d114a1e00262542'

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        self.metadata_url = u'https://drive.amazonaws.com/drive/v1/'
        self.content_url = u'https://content-na.drive.amazonaws.com/cdproxy/'

        self.names_to_ids = {}
        self.backup_target_id = None
        self.backup_target = parsed_url.path.lstrip(u'/')

        if globals.volsize > (10 * 1024 * 1024 * 1024):
            # https://forums.developer.amazon.com/questions/22713/file-size-limits.html
            # https://forums.developer.amazon.com/questions/22038/support-for-chunked-transfer-encoding.html
            log.FatalError(
                u'Your --volsize is bigger than 10 GiB, which is the maximum '
                u'file size on Amazon Drive that does not require work arounds.')

        try:
            global requests
            global OAuth2Session
            import requests
            from requests_oauthlib import OAuth2Session
        except ImportError:
            raise BackendException(
                u'Amazon Drive backend requires python-requests and '
                u'python-requests-oauthlib to be installed.\n\n'
                u'For Debian and derivates use:\n'
                u'  apt-get install python-requests python-requests-oauthlib\n'
                u'For Fedora and derivates use:\n'
                u'  yum install python-requests python-requests-oauthlib')

        self.initialize_oauth2_session()
        self.resolve_backup_target()

    def initialize_oauth2_session(self):
        u"""Setup or refresh oauth2 session with Amazon Drive"""

        def token_updater(token):
            u"""Stores oauth2 token on disk"""
            try:
                with open(self.OAUTH_TOKEN_PATH, u'w') as f:
                    json.dump(token, f)
            except Exception as err:
                log.Error(u'Could not save the OAuth2 token to %s. This means '
                          u'you may need to do the OAuth2 authorization '
                          u'process again soon. Original error: %s' % (
                              self.OAUTH_TOKEN_PATH, err))

        token = None
        try:
            with open(self.OAUTH_TOKEN_PATH) as f:
                token = json.load(f)
        except IOError as err:
            log.Notice(u'Could not load OAuth2 token. '
                       u'Trying to create a new one. (original error: %s)' % err)

        self.http_client = OAuth2Session(
            self.CLIENT_ID,
            scope=self.OAUTH_SCOPE,
            redirect_uri=self.OAUTH_REDIRECT_URL,
            token=token,
            auto_refresh_kwargs={
                u'client_id': self.CLIENT_ID,
                u'client_secret': self.CLIENT_SECRET,
            },
            auto_refresh_url=self.OAUTH_TOKEN_URL,
            token_updater=token_updater)

        if token is not None:
            self.http_client.refresh_token(self.OAUTH_TOKEN_URL)

        endpoints_response = self.http_client.get(self.metadata_url +
                                                  u'account/endpoint')
        if endpoints_response.status_code != requests.codes.ok:
            token = None

        if token is None:
            if not sys.stdout.isatty() or not sys.stdin.isatty():
                log.FatalError(u'The OAuth2 token could not be loaded from %s '
                               u'and you are not running duplicity '
                               u'interactively, so duplicity cannot possibly '
                               u'access Amazon Drive.' % self.OAUTH_TOKEN_PATH)
            authorization_url, _ = self.http_client.authorization_url(
                self.OAUTH_AUTHORIZE_URL)

            print(u'')
            print(u'In order to allow duplicity to access Amazon Drive, please '
                  u'open the following URL in a browser and copy the URL of the '
                  u'page you see after authorization here:')
            print(authorization_url)
            print(u'')

            redirected_to = (input(u'URL of the resulting page: ')
                             .replace(u'http://', u'https://', 1)).strip()

            token = self.http_client.fetch_token(
                self.OAUTH_TOKEN_URL,
                client_secret=self.CLIENT_SECRET,
                authorization_response=redirected_to)

            endpoints_response = self.http_client.get(self.metadata_url +
                                                      u'account/endpoint')
            endpoints_response.raise_for_status()
            token_updater(token)

        urls = endpoints_response.json()
        if u'metadataUrl' not in urls or u'contentUrl' not in urls:
            log.FatalError(u'Could not retrieve endpoint URLs for this account')
        self.metadata_url = urls[u'metadataUrl']
        self.content_url = urls[u'contentUrl']

    def resolve_backup_target(self):
        u"""Resolve node id for remote backup target folder"""

        response = self.http_client.get(
            self.metadata_url + u'nodes?filters=kind:FOLDER AND isRoot:true')
        parent_node_id = response.json()[u'data'][0][u'id']

        for component in [x for x in self.backup_target.split(u'/') if x]:
            # There doesn't seem to be escaping support, so cut off filter
            # after first unsupported character
            query = re.search(u'^[A-Za-z0-9_-]*', component).group(0)
            if component != query:
                query = query + u'*'

            matches = self.read_all_pages(
                self.metadata_url + u'nodes?filters=kind:FOLDER AND name:%s '
                                    u'AND parents:%s' % (query, parent_node_id))
            candidates = [f for f in matches if f.get(u'name') == component]

            if len(candidates) >= 2:
                log.FatalError(u'There are multiple folders with the same name '
                               u'below one parent.\nParentID: %s\nFolderName: '
                               u'%s' % (parent_node_id, component))
            elif len(candidates) == 1:
                parent_node_id = candidates[0][u'id']
            else:
                log.Debug(u'Folder %s does not exist yet. Creating.' % component)
                parent_node_id = self.mkdir(parent_node_id, component)

        log.Debug(u"Backup target folder has id: %s" % parent_node_id)
        self.backup_target_id = parent_node_id

    def get_file_id(self, remote_filename):
        u"""Find id of remote file in backup target folder"""

        if remote_filename not in self.names_to_ids:
            self._list()

        return self.names_to_ids.get(remote_filename)

    def mkdir(self, parent_node_id, folder_name):
        u"""Create a new folder as a child of a parent node"""

        data = {u'name': folder_name, u'parents': [parent_node_id], u'kind': u'FOLDER'}
        response = self.http_client.post(
            self.metadata_url + u'nodes',
            data=json.dumps(data))
        response.raise_for_status()
        return response.json()[u'id']

    def multipart_stream(self, metadata, source_path):
        u"""Generator for multipart/form-data file upload from source file"""

        boundary = self.MULTIPART_BOUNDARY

        yield str.encode(u'--%s\r\nContent-Disposition: form-data; '
                         u'name="metadata"\r\n\r\n' % boundary +
                         u'%s\r\n' % json.dumps(metadata) +
                         u'--%s\r\n' % boundary)
        yield b'Content-Disposition: form-data; name="content"; filename="i_love_backups"\r\n'
        yield b'Content-Type: application/octet-stream\r\n\r\n'

        with source_path.open() as stream:
            while True:
                f = stream.read(DEFAULT_BUFFER_SIZE)
                if f:
                    yield f
                else:
                    break

        yield str.encode(u'\r\n--%s--\r\n' % boundary +
                         u'multipart/form-data; boundary=%s' % boundary)

    def read_all_pages(self, url):
        u"""Iterates over nodes API URL until all pages were read"""

        result = []
        next_token = u''
        token_param = u'&startToken=' if u'?' in url else u'?startToken='

        while True:
            paginated_url = url + token_param + next_token
            response = self.http_client.get(paginated_url)
            if response.status_code != 200:
                raise BackendException(u"Pagination failed with status=%s on "
                                       u"URL=%s" % (response.status_code, url))

            parsed = response.json()
            if u'data' in parsed and len(parsed[u'data']) > 0:
                result.extend(parsed[u'data'])
            else:
                break

            # Do not make another HTTP request if everything is here already
            if len(result) >= parsed[u'count']:
                break

            if u'nextToken' not in parsed:
                break
            next_token = parsed[u'nextToken']

        return result

    def raise_for_existing_file(self, remote_filename):
        u"""Report error when file already existed in location and delete it"""

        self._delete(remote_filename)
        raise BackendException(u'Upload failed, because there was a file with '
                               u'the same name as %s already present. The file was '
                               u'deleted, and duplicity will retry the upload unless '
                               u'the retry limit has been reached.' % remote_filename)

    def _put(self, source_path, remote_filename):
        u"""Upload a local file to Amazon Drive"""

        quota = self.http_client.get(self.metadata_url + u'account/quota')
        quota.raise_for_status()
        available = quota.json()[u'available']

        source_size = os.path.getsize(source_path.name)

        if source_size > available:
            raise BackendException(
                u'Out of space: trying to store "%s" (%d bytes), but only '
                u'%d bytes available on Amazon Drive.' % (
                    source_path.name, source_size, available))

        # Just check the cached list, to avoid _list for every new file being
        # uploaded
        if remote_filename in self.names_to_ids:
            log.Debug(u'File %s seems to already exist on Amazon Drive. Deleting '
                      u'before attempting to upload it again.' % remote_filename)
            self._delete(remote_filename)

        metadata = {u'name': remote_filename, u'kind': u'FILE',
                    u'parents': [self.backup_target_id]}
        headers = {u'Content-Type': u'multipart/form-data; boundary=%s' % self.MULTIPART_BOUNDARY}
        data = self.multipart_stream(metadata, source_path)

        response = self.http_client.post(
            self.content_url + u'nodes?suppress=deduplication',
            data=data,
            headers=headers)

        if response.status_code == 409:  # "409 : Duplicate file exists."
            self.raise_for_existing_file(remote_filename)
        elif response.status_code == 201:
            log.Debug(u'%s uploaded successfully' % remote_filename)
        elif response.status_code == 408 or response.status_code == 504:
            log.Info(u'%s upload failed with timeout status code=%d. Speculatively '
                     u'waiting for %d seconds to see if Amazon Drive finished the '
                     u'upload anyway' % (remote_filename, response.status_code,
                                         globals.timeout))
            tries = globals.timeout / 15
            while tries >= 0:
                tries -= 1
                time.sleep(15)

                remote_size = self._query(remote_filename)[u'size']
                if source_size == remote_size:
                    log.Debug(u'Upload turned out to be successful after all.')
                    return
                elif remote_size == -1:
                    log.Debug(u'Uploaded file is not yet there, %d tries left.'
                              % (tries + 1))
                    continue
                else:
                    self.raise_for_existing_file(remote_filename)
            raise BackendException(u'%s upload failed and file did not show up '
                                   u'within time limit.' % remote_filename)
        else:
            log.Debug(u'%s upload returned an undesirable status code %s'
                      % (remote_filename, response.status_code))
            response.raise_for_status()

        parsed = response.json()
        if u'id' not in parsed:
            raise BackendException(u'%s was uploaded, but returned JSON does not '
                                   u'contain ID of new file. Retrying.\nJSON:\n\n%s'
                                   % (remote_filename, parsed))

        # XXX: The upload may be considered finished before the file shows up
        # in the file listing. As such, the following is required to avoid race
        # conditions when duplicity calls _query or _list.
        self.names_to_ids[parsed[u'name']] = parsed[u'id']

    def _get(self, remote_filename, local_path):
        u"""Download file from Amazon Drive"""

        with local_path.open(u'wb') as local_file:
            file_id = self.get_file_id(remote_filename)
            if file_id is None:
                raise BackendException(
                    u'File "%s" cannot be downloaded: it does not exist' %
                    remote_filename)

            response = self.http_client.get(
                self.content_url + u'/nodes/' + file_id + u'/content', stream=True)
            response.raise_for_status()
            for chunk in response.iter_content(chunk_size=DEFAULT_BUFFER_SIZE):
                if chunk:
                    local_file.write(chunk)
            local_file.flush()

    def _query(self, remote_filename):
        u"""Retrieve file size info from Amazon Drive"""

        file_id = self.get_file_id(remote_filename)
        if file_id is None:
            return {u'size': -1}
        response = self.http_client.get(self.metadata_url + u'nodes/' + file_id)
        response.raise_for_status()

        return {u'size': response.json()[u'contentProperties'][u'size']}

    def _list(self):
        u"""List files in Amazon Drive backup folder"""

        files = self.read_all_pages(
            self.metadata_url + u'nodes/' + self.backup_target_id +
            u'/children?filters=kind:FILE')

        self.names_to_ids = {f[u'name']: f[u'id'] for f in files}

        return list(self.names_to_ids.keys())

    def _delete(self, remote_filename):
        u"""Delete file from Amazon Drive"""

        file_id = self.get_file_id(remote_filename)
        if file_id is None:
            raise BackendException(
                u'File "%s" cannot be deleted: it does not exist' % (
                    remote_filename))
        response = self.http_client.put(self.metadata_url + u'trash/' + file_id)
        response.raise_for_status()
        del self.names_to_ids[remote_filename]


duplicity.backend.register_backend(u'ad', ADBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os

import duplicity.backend
from duplicity import log
from duplicity import path
from duplicity.errors import BackendException


class LocalBackend(duplicity.backend.Backend):
    u"""Use this backend when saving to local disk

    Urls look like file://testfiles/output.  Relative to root can be
    gotten with extra slash (file:///usr/local).

    """
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        # The URL form "file:MyFile" is not a valid duplicity target.
        if not parsed_url.path.startswith(u'//'):
            raise BackendException(u"Bad file:// path syntax.")
        self.remote_pathdir = path.Path(parsed_url.path[2:])
        try:
            os.makedirs(self.remote_pathdir.base)
        except Exception:
            pass

    def _move(self, source_path, remote_filename):
        target_path = self.remote_pathdir.append(remote_filename)
        try:
            source_path.rename(target_path)
            return True
        except OSError:
            return False

    def _put(self, source_path, remote_filename):
        target_path = self.remote_pathdir.append(remote_filename)
        target_path.writefileobj(source_path.open(u"rb"))

    def _get(self, filename, local_path):
        source_path = self.remote_pathdir.append(filename)
        local_path.writefileobj(source_path.open(u"rb"))

    def _list(self):
        return self.remote_pathdir.listdir()

    def _delete(self, filename):
        self.remote_pathdir.append(filename).delete()

    def _query(self, filename):
        target_file = self.remote_pathdir.append(filename)
        target_file.setdata()
        size = target_file.getsize() if target_file.exists() else -1
        return {u'size': size}


duplicity.backend.register_backend(u"file", LocalBackend)
